package net.afyre.haha.demo.solver.smt;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;
import java.util.concurrent.TimeUnit;

import net.afyre.haha.demo.solver.context.Answer;
import net.afyre.haha.demo.solver.context.Declaration;
import net.afyre.haha.demo.solver.context.DeclarationVisitor;
import net.afyre.haha.demo.solver.context.SolverContext;
import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.formulae.impl.Negation;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.types.ValueType;
import net.afyre.haha.demo.solver.smt.sexpr.DefaultSExprFactory;
import net.afyre.haha.demo.solver.smt.sexpr.SExpr;
import net.afyre.haha.demo.solver.smt.sexpr.SExprFactory;
import net.afyre.haha.demo.util.EmptyIterable;

/**
 * Solver interface implementation.
 * 
 * This only works with Z3 for now.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 */
public abstract class SmtSolver implements SolverContext {
  private static final int EXIT_TIMEOUT = 3000;
  private static final int REPLY_TIMEOUT = 15000;
  
  /** Handle used for external process management. */
  private final SolverProcess solver;

  /** Used to print errors and warnings. */
  private final SmtLogger logger;
  
  /** Flag active during a restart after an error. */
  private boolean restarting;
  
  private final SExprFactory exprFactory;
  private final SExprBuilder exprBuilder;

  private final DeclarationProcessor declarationProcessor;

  private final SExpr[] checkSat;
  
  private final Stack<List<Declaration>> declarations = 
      new Stack<List<Declaration>>();; 
  private int stackLevel;
  
  /* TODO: Add this to generic solver interface. */
  /** Last counterexample found. */
  private SmtModel lastModel;
  
  /**
   * Constructor.
   * @param solverPath Path to the Z3 executable.
   * @param solverArgs Command-line arguments.
   * @param logger Object used to print log messages.
   */
  protected SmtSolver(final String solverPath,
                      final String[] solverArgs,
                      final SmtLogger logger) {
    this.solver = new SolverProcess(logger, solverPath, solverArgs);
    this.logger = logger;
    restarting = false;
    declarations.push(new LinkedList<Declaration>());
    exprFactory = new DefaultSExprFactory();
    exprBuilder = new SExprBuilder();
    declarationProcessor = new DeclarationProcessor();
    stackLevel = 0;
    checkSat = new SExpr[] {
       apply("check-sat"),
       /*apply("check-sat-using", 
         apply("then", apply("using-params", 
                 exprFactory.createSymbol("simplify"),
                 exprFactory.createKeyword("arith-lhs"),
                 exprFactory.createSymbol("true"),
                 exprFactory.createKeyword("som"),
                 exprFactory.createSymbol("true")),
               exprFactory.createSymbol("snf"), 
               exprFactory.createSymbol("solve-eqs"), 
               exprFactory.createSymbol("smt"))) */};
  }
  
  /**
   * Get last counterexample.
   * @return Last counterexample.
   */
  public SmtModel getLastModel() {
    return lastModel;
  }

  /**
   * Send an expression to the solver.
   * @param expr Expression to be sent.
   * @throws CancellationException  If I/O was interrupted.
   */
  protected void send(final SExpr expr) throws CancellationException {
    if (Thread.interrupted()) {
      Thread.currentThread().interrupt();
      throw new CancellationException();
    }
    if (!solver.isRunning()) {
      solver.start();
      sendPreamble();
    }
    if (!solver.send(expr)) {
      if (Thread.interrupted()) {
        Thread.currentThread().interrupt();
        throw new CancellationException();
      }
      if (!restarting) {
        restarting = true;
        try {
          restartSolver();
        } finally {
          restarting = false;
        }
      }
      throw new CancellationException();
    }
  }

  private void restoreContext() throws CancellationException {
    //send(apply("reset"));
    for (int i = 0; i <= stackLevel; i++) {
      if (i > 0) { 
        send(apply("push"));
      }
      for (final Declaration decl : declarations.get(i)) {
        try {
          decl.accept(declarationProcessor);
        } catch (final WrappedCancellationException e) {
          throw e.getException();
        }
      }
    }
  }
  
  private void restartSolver() throws CancellationException {
    assert restarting;
    
    solver.stop(EXIT_TIMEOUT, TimeUnit.MILLISECONDS);
    restoreContext();
  }
  
  /**
   * Invoked after solver start or restart.
   * @throws CancellationException If I/O was interrupted.
   */
  protected abstract void sendPreamble() throws CancellationException;
  
  /**
   * Send a command to set a named SMT option.
   * @param name Option name.
   * @param value Option value.
   * @throws CancellationException If I/O was interrupted.
   */
  protected void sendOption(final String name, final String value) 
    throws CancellationException {
    
    final SExpr cmdExpr = exprFactory.createSymbol("set-option");
    final SExpr nameExpr = exprFactory.createKeyword(name);
    final SExpr valueExpr = exprFactory.createSymbol(value);

    final SExpr expr = exprFactory.createList(
                         Arrays.asList(cmdExpr, nameExpr, valueExpr));
    send(expr);
  }
  
  /**
   * Create an expression from a function name and arguments.
   * @param op Function name (will be converted to a symbol).
   * @param args Arguments (as S-Expressions).
   * @return S-Expression.
   */
  protected SExpr apply(final String op, 
                        final SExpr... args) {
    final List<SExpr> elements = new ArrayList<SExpr>();
    elements.add(exprFactory.createSymbol(op));
    for (final SExpr argument : args) {
      elements.add(argument);
    }
    return exprFactory.createList(elements);
  }

  /**
   * Convert a list of strings to a S-Expression.
   * @param args Strings.
   * @return S-Expression.
   */
  protected SExpr list(final String... args) {
    final List<SExpr> elements = new ArrayList<SExpr>();
    for (final String argument : args) {
      elements.add(exprFactory.createSymbol(argument));
    }
    return exprFactory.createList(elements);
  }
  
  /**
   * Read an expression printed by the solver.
   * @return Expression or null (error).
   */
  protected SExpr read() {
    return solver.read(REPLY_TIMEOUT, TimeUnit.MILLISECONDS);
  }
  
  /** {@inheritDoc} */
  @Override
  public boolean addDeclaration(
      final Declaration declaration) {
    declarations.get(stackLevel).add(declaration);
    try {
      declaration.accept(declarationProcessor);
    } catch (final WrappedCancellationException e) {
      return false;
    }
    return true;
  }

  /** {@inheritDoc} */
  @Override
  public boolean mark() {
    try {
      doMark();
    } catch (final CancellationException e) {
      return false;
    }
    return true;
  }
  
  private void doMark() throws CancellationException {
    stackLevel++;
    declarations.push(new LinkedList<Declaration>());
    send(apply("push"));
  }

  /** {@inheritDoc} */
  @Override
  public boolean reset() {
    try {
      doReset();
    } catch (final CancellationException e) {
      return false;
    }
    return true;
  }
  
  private void doReset() throws CancellationException {
    if (stackLevel == 0) {
      send(apply("reset"));
      declarations.set(0, new LinkedList<Declaration>());
    } else {
      stackLevel--;
      send(apply("pop"));
      declarations.pop();
    }
  }
  
  private Answer parseResult(final SExpr result) {
    if (result == null) {
      return Answer.ERROR;
    }
    final String text = result.toString().trim().toLowerCase();
    final Answer answer;
    if ("sat".equals(text)) {
      answer = Answer.VALID;
    } else if ("unsat".equals(text)) {
      answer = Answer.INVALID;
    } else {
      answer = Answer.UNKNOWN;
    }
    return answer;
  }
  
  /** {@inheritDoc} */
  @Override
  public Answer isValid(final Formula query) {
    return isSatisfiable(new Negation(query)).neg();
  }

  /** {@inheritDoc} */
  @Override
  public Answer isSatisfiable(final Formula query) {
    try {
      return checkSat(query);
    } catch (final CancellationException e) {
      return Answer.ERROR;
    }
  }
  
  private Answer checkSat(final Formula query) throws CancellationException {
    Answer result = Answer.UNKNOWN;
    
    lastModel = null;
    mark();
    send(apply("assert", query.accept(exprBuilder)));
    for (final SExpr cmd : checkSat) {
      send(cmd);
      final SExpr resultExpr = 
          solver.read(REPLY_TIMEOUT, TimeUnit.MILLISECONDS);
      result = parseResult(resultExpr);
      if (result != Answer.UNKNOWN) { break; }
    }
    
    switch (result) {
      case VALID:
        produceModel();
        reset();
        break;
      case INVALID:
        reset();
        break;
      default:
        reset();
        break;
    }
    return result;
  }

  /** {@inheritDoc} */
  @Override
  public void finish() {
    solver.send(apply("exit"));
    solver.stop(EXIT_TIMEOUT, TimeUnit.MILLISECONDS);
  }

  /** Get model (counterexample) from the solver and report it. */
  private void produceModel() throws CancellationException {
    send(apply("get-model"));
    final SExpr model = solver.read(REPLY_TIMEOUT, TimeUnit.MILLISECONDS);
    if (model != null) {
      try {
        lastModel = SmtModel.fromGetModel(model);
        logger.model(lastModel);
      } catch (final IllegalArgumentException e) {
        logger.warning("Unable to parse the counterexample", e);
      }
    }    
  }

  private void wrappedSend(final SExpr expr) {
    try {
      send(expr);
    } catch (final CancellationException e) {
      throw new WrappedCancellationException(e);
    }
  }
  
  private final class DeclarationProcessor implements DeclarationVisitor {
    /** {@inheritDoc} */
    @Override
    public void visitAxiom(final String name, 
                           final Formula axiom) {
      wrappedSend(apply("assert", axiom.accept(exprBuilder)));
    }

    /** {@inheritDoc} */
    @Override
    public void visitTermDeclaration(final String name, 
                                     final ValueType type) {
      wrappedSend(apply(
         "declare-fun", 
         exprFactory.createSymbol(name), 
         exprFactory.createList(EmptyIterable.<SExpr>getInstance()),
         type.accept(exprBuilder)));
    }

    /** {@inheritDoc} */
    @Override
    public void visitTermDefinition(final String name, 
                                    final ValueType type, 
                                    final Term value) {
      final SExpr typeExpr = type.accept(exprBuilder);
      final SExpr valueExpr = value.accept(exprBuilder);
      
      wrappedSend(apply(
         "define-fun", 
         exprFactory.createSymbol(name), 
         exprFactory.createList(EmptyIterable.<SExpr>getInstance()),
         typeExpr, valueExpr));
    }

    /** {@inheritDoc} */
    @Override
    public void visitPropDeclaration(final String name) {
      wrappedSend(apply(
         "declare-fun", 
         exprFactory.createSymbol(name), 
         exprFactory.createList(EmptyIterable.<SExpr>getInstance()),
         apply("Bool")));
    }

    /** {@inheritDoc} */
    @Override
    public void visitPropDefinition(final String name, 
                                    final Formula value) {
      final SExpr typeExpr = apply("Bool");

      final SExpr valueExpr = value.accept(exprBuilder);
      
      wrappedSend(apply(
         "define-fun", 
         exprFactory.createSymbol(name), 
         exprFactory.createList(EmptyIterable.<SExpr>getInstance()),
         typeExpr, valueExpr));
    }

    /** {@inheritDoc} */
    @Override
    public void visitFunctionDeclaration(
        final String name,
        final Iterable<VariableBinding> args,
        final ValueType resultType) {
      final SExpr nameExpr = exprFactory.createSymbol(name);      
      
      final List<SExpr> argExprs = new ArrayList<SExpr>();
      for (final VariableBinding arg : args) {
        argExprs.add(arg.getType().accept(exprBuilder));
      }
      
      final SExpr resultExpr = resultType.accept(exprBuilder);
      
      wrappedSend(apply("declare-fun", 
                 nameExpr,
                 exprFactory.createList(argExprs),
                 resultExpr));
    }

    /** {@inheritDoc} */
    @Override
    public void visitFunctionDefinition(
        final String name,
        final Iterable<VariableBinding> args,
        final ValueType resultType, 
        final Term body) {
      final SExpr nameExpr = exprFactory.createSymbol(name);      
      
      final List<SExpr> argExprs = new ArrayList<SExpr>();
      for (final VariableBinding arg : args) {
        final SExpr argNameExpr = 
            exprFactory.createSymbol(arg.getName());
        final SExpr argTypeExpr = arg.getType().accept(exprBuilder);
        final List<SExpr> bindings = new LinkedList<SExpr>();
        bindings.add(argNameExpr);
        bindings.add(argTypeExpr);
        argExprs.add(exprFactory.createList(bindings));
      }
      
      final SExpr resultExpr = resultType.accept(exprBuilder);
      
      final SExpr bodyExpr = body.accept(exprBuilder);
      
      wrappedSend(apply("define-fun", 
                 nameExpr,
                 exprFactory.createList(argExprs),
                 resultExpr, 
                 bodyExpr));
    }

    /** {@inheritDoc} */
    @Override
    public void visitPredicateDeclaration(
        final String name,
        final Iterable<VariableBinding> args) {
      final SExpr nameExpr = exprFactory.createSymbol(name);      
      
      final List<SExpr> argExprs = new ArrayList<SExpr>();
      for (final VariableBinding arg : args) {
        argExprs.add(arg.getType().accept(exprBuilder));
      }
      
      final SExpr resultExpr = apply("Bool");
      
      wrappedSend(apply("declare-fun", 
                 nameExpr,
                 exprFactory.createList(argExprs),
                 resultExpr));
    }

    /** {@inheritDoc} */
    @Override
    public void visitPredicateDefinition(
        final String name,
        final Iterable<VariableBinding> args,
        final Formula body) {
      final SExpr nameExpr = exprFactory.createSymbol(name);      
      
      final List<SExpr> argExprs = new ArrayList<SExpr>();
      for (final VariableBinding arg : args) {
        final SExpr argNameExpr = 
            exprFactory.createSymbol(arg.getName());
        final SExpr argTypeExpr = arg.getType().accept(exprBuilder);
        final List<SExpr> bindings = new LinkedList<SExpr>();
        bindings.add(argNameExpr);
        bindings.add(argTypeExpr);
        argExprs.add(exprFactory.createList(bindings));
      }
      
      final SExpr resultExpr = exprFactory.createSymbol("Bool");     
      final SExpr bodyExpr = body.accept(exprBuilder);
      
      wrappedSend(apply("define-fun", 
                 nameExpr,
                 exprFactory.createList(argExprs),
                 resultExpr, 
                 bodyExpr));
    }
  }
  
  /**
   * Exception used to signal that execution of a command was interrupted.
   */
  protected static final class CancellationException extends Exception {
    private static final long serialVersionUID = 1L;   
  }
  
  private static final class WrappedCancellationException 
    extends RuntimeException {
    
    private static final long serialVersionUID = 1L;
    private final CancellationException exception;
    
    public WrappedCancellationException(final CancellationException cause) {
      super(cause);
      exception = cause;
    }
    
    public CancellationException getException() {
      return exception;
    }
  }
}
