package net.afyre.haha.demo.solver.smt.sexpr;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.regex.Pattern;

/**
 * Default implementation of a S-Expression factory.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 */
public class DefaultSExprFactory implements SExprFactory {
  /** Pattern describing valid keywords and unquoted symbols. */
  private static final Pattern VALID_KEYWORD = 
      Pattern.compile("\\A[-\\p{Alpha}~!@$%^&*_+=<>.?/]" + 
                      "[-\\p{Alnum}~!@$%^&*_+=<>.?/]*\\z");
  

  /** {@inheritDoc} */
  @Override
  public SExpr createInteger(final BigInteger value) {
    return new IntegerExpr(value);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr createDecimal(final BigDecimal value) {
    return new DecimalExpr(value);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr createHex(final BigInteger value, final int length) {
    return new HexExpr(value, length);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr createBin(final BigInteger value, final int length) {
    return new BinExpr(value, length);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr createString(final String value) {
    return new StringExpr(value);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr createSymbol(final String name) {
    return new SymbolExpr(name);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr createKeyword(final String name) {
    return new KeywordExpr(name);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr createList(final Iterable<? extends SExpr> elements) {
    return new ListExpr(elements);
  }  
  
  /** 
   * Base class for concrete implementations of expressions.
   * Provied default implementation of toString().
   */
  private abstract static class Expr implements SExpr {
    /** {@inheritDoc} */
    @Override
    public String toString() {
      return doSwitch(SExprPrinter.getInstance());
    }
  }
  
  /** Concrete implementation of integer numerals. */
  private static final class IntegerExpr extends Expr {
    /** Integer value. */
    private final BigInteger value;
    
    /**
     * Constructor.
     * @param value Integer value.
     * @throws IllegalArgumentException If the value is negative.
     * @throw NullPointerException If the value is null.
     */
    public IntegerExpr(final BigInteger value) {
      if (value == null) { throw new NullPointerException("value"); }
      if (value.signum() == -1) {
        throw new IllegalArgumentException("value");
      }
      this.value = value;
    }
    
    /** {@inheritDoc} */
    @Override
    public <T> T doSwitch(final SExprSwitch<T> sw) {
      return sw.caseInteger(value);
    }
  }
  
  /** Concrete implementation of decimal numerals. */
  private static final class DecimalExpr extends Expr {
    /** Integer value. */
    private final BigDecimal value;
    
    /**
     * Constructor.
     * @param value Decimal value.
     * @throws IllegalArgumentException If the value is negative.
     * @throw NullPointerException If the value is null.
     */
    public DecimalExpr(final BigDecimal value) {
      if (value == null) { throw new NullPointerException("value"); }
      if (value.signum() == -1) {
        throw new IllegalArgumentException("value");
      }
      this.value = value;
    }
    
    /** {@inheritDoc} */
    @Override
    public <T> T doSwitch(final SExprSwitch<T> sw) {
      if (value == null) { throw new NullPointerException("value"); }
      return sw.caseDecimal(value);
    }
  }

  /** Concrete implementation of hex numerals. */
  private static final class HexExpr extends Expr {
    /** Numeral value. */
    private final BigInteger value;
    
    /** Number of digits. */
    private final int length;
    
    /**
     * Constructor.
     * @param value Numeral value.
     * @throws IllegalArgumentException If the value is negative.
     * @throw NullPointerException If the value is null.
     */
    public HexExpr(final BigInteger value, final int length) {
      if (value == null) { throw new NullPointerException("value"); }
      if (value.signum() == -1) {
        throw new IllegalArgumentException("value");
      }
      this.value = value;
      this.length = length;
    }
    
    /** {@inheritDoc} */
    @Override
    public <T> T doSwitch(final SExprSwitch<T> sw) {
      return sw.caseHex(value, length);
    }
  }
 
  /** Concrete implementation of binary numerals. */
  private static final class BinExpr extends Expr {
    /** Numeral value. */
    private final BigInteger value;
    
    /** Number of digits. */
    private final int length;
    
    /**
     * Constructor.
     * @param value Numeral value.
     * @throws IllegalArgumentException If the value is negative.
     * @throw NullPointerException If the value is null.
     */
    public BinExpr(final BigInteger value, final int length) {
      if (value == null) { throw new NullPointerException("value"); }
      if (value.signum() == -1) {
        throw new IllegalArgumentException("value");
      }
      this.value = value;
      this.length = length;
    }
    
    /** {@inheritDoc} */
    @Override
    public <T> T doSwitch(final SExprSwitch<T> sw) {
      return sw.caseBin(value, length);
    }
  }
  
  /** Concrete implementation of quoted strings. */
  private static final class StringExpr extends Expr {  
    /** String value. */
    private final String value;
    
    /**
     * Constructor.
     * @param value String value.
     * @throw NullPointerException If the value is null.
     */
    public StringExpr(final String value) {
      if (value == null) { throw new NullPointerException("value"); }
      this.value = value;
    }
    
    /** {@inheritDoc} */
    @Override
    public <T> T doSwitch(final SExprSwitch<T> sw) {
      return sw.caseString(value);
    }
  }
  
  /** Concrete implementation of symbols. */
  private static final class SymbolExpr extends Expr {  
    /** Symbol name. */
    private final String name;
    
    /**
     * Constructor.
     * @param name Symbol name.
     * @throw NullPointerException If the name is null.
     */
    public SymbolExpr(final String name) {
      if (name == null) { throw new NullPointerException("name"); }
      if (name.contains("\\") || name.contains("|")) {
        throw new IllegalArgumentException("name");
      }
      this.name = name;
    }
    
    /** {@inheritDoc} */
    @Override
    public <T> T doSwitch(final SExprSwitch<T> sw) {
      return sw.caseSymbol(name);
    }
  }
  
  /** Concrete implementation of keywords. */
  private static final class KeywordExpr extends Expr {  
    /** Keyword name. */
    private final String name;
    
    /**
     * Constructor.
     * @param name Keyword name.
     * @throw NullPointerException If the name is null.
     */
    public KeywordExpr(final String name) {
      if (name == null) { throw new NullPointerException("name"); }
      if (!VALID_KEYWORD.matcher(name).matches()) {
        throw new IllegalArgumentException("name");
      }
      this.name = name;
    }
    
    /** {@inheritDoc} */
    @Override
    public <T> T doSwitch(final SExprSwitch<T> sw) {
      return sw.caseKeyword(name);
    }
  }
  
  /** Concrete implementation of lists. */
  private static final class ListExpr extends Expr {
    /** Elements. */
    private final List<SExpr> elements;
    
    /**
     * Constructor.
     * @param elements Elements.
     * @throw NullPointerException If the list of elements is null.
     */
    public ListExpr(final Iterable<? extends SExpr> elements) {
      if (elements == null) { 
        throw new NullPointerException("elements"); 
      }
      this.elements = new ArrayList<SExpr>();
      for (final SExpr elem : elements) {
        this.elements.add(elem);
      }
    }
    
    /** {@inheritDoc} */
    @Override
    public <T> T doSwitch(final SExprSwitch<T> sw) {
      return sw.caseList(Collections.unmodifiableList(elements));
    } 
  }
}
