package net.afyre.haha.demo.solver.smt.sexpr;

import java.math.BigDecimal;
import java.math.BigInteger;

/**
 * Interface of objects used to create S-Expressions.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public interface SExprFactory {
  /**
   * Create an integer numeral.
   * @param value Integer value (non-negative).
   * @return Numeral object.
   * @throws IllegalArgumentException If the value is negative.
   * @throw NullPointerException If the value is null. 
   */
  SExpr createInteger(BigInteger value);
  
  /**
   * Create a decimal numeral.
   * @param value Numeral value (non-negative).
   * @return Numeral object.
   * @throws IllegalArgumentException If the value is negative.
   * @throw NullPointerException If the value is null.
   */
  SExpr createDecimal(final BigDecimal value);
  
  /**
   * Create a hexadecimal numeral.
   * @param value Numeral value (non-negative).
   * @param length Minimal number of digits.
   *               Can be used to introduce leading zeros. 
   * @return Numeral object.
   * @throws IllegalArgumentException If the value is negative.
   * @throw NullPointerException If the value is null.
   */
  SExpr createHex(BigInteger value, int length);
  

  /**
   * Create a binary numeral.
   * @param value Numeral value (non-negative).
   * @param length Minimal number of digits.
   *               Can be used to introduce leading zeros. 
   * @return Numeral object.
   * @throws IllegalArgumentException If the value is negative.
   * @throw NullPointerException If the value is null.
   */
  SExpr createBin(BigInteger value, int length);
  
  /**
   * Create a quoted string.
   * @param value String text.
   * @return String object.
   * @throw NullPointerException If the value is null.
   */
  SExpr createString(String value);
  
  /**
   * Create a symbol.
   * @param name Symbol name.
   * @return Symbol object.
   * @throws IllegalArgumentException If the symbol name is not valid.
   * @throw NullPointerException If the symbol name is null.
   */
  SExpr createSymbol(String name);
  
  /**
   * Create a keyword.
   * @param name Keyword text (without leading colon).
   * @return Symbol object.
   * @throws IllegalArgumentException If the keyword text is not valid.
   * @throw NullPointerException If the keyword text is null.
   */
  SExpr createKeyword(String name);
    
  /**
   * Create a list.
   * @param elements List elements.
   * @return List object.
   * @throw NullPointerException If the list of elements is null.
   */
  SExpr createList(Iterable<? extends SExpr> elements);
}
