package net.afyre.haha.demo.solver.smt.sexpr;

import java.io.IOException;
import java.io.Reader;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.LinkedList;
import java.util.List;

/**
 * Class used to parse S-Expressions.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public class SExprParser {
  /** Characters usable in unquoted identifiers. */
  private static final String OPERATOR_CHARS = "~!@$%^&*_+-=<>.?/";
  
  /** Factory used to build S-Expressions. */
  private final SExprFactory factory;
  
  /**
   * Constructor.
   * @param factory Factory which will be used to create S-Expressions.
   */
  public SExprParser(final SExprFactory factory) {
    this.factory = factory;
  }
  
  
  /**
   * Read an expression from an input stream.
   * 
   * @param input Input stream (must allow pushback).
   * @return Parsed expression.
   * @throws ParserException In case of a syntax or I/O error.
   */
  public SExpr parse(final Reader input) {
    int firstChar;
    
    if (!input.markSupported()) {
      throw new ParserException(
        "Input stream does not support pushback.");
    }
    
    skipWhitespace(input);
    firstChar = peek(input);
    final SExpr result;
    if (firstChar == -1) {
      throw new ParserException("Unexpected end of stream");
    } else if (firstChar == '(') {
      result = parseList(input);
    } else if (isSymbolStart(firstChar)) {
      result = parseSymbol(input);
    } else if (Character.isDigit(firstChar)) {
      result = parseDec(input);
    } else if (firstChar == '#') {
      result = parseHexOrBin(input);
    } else if (firstChar == ':') {
      result = parseKeyword(input);
    } else if (firstChar == '"') {
      result = parseString(input);
    } else if (firstChar == '|') {
      result = parseQuotedSymbol(input);
    } else {
      throw new ParserException(
          String.format("Invalid character (%c)", firstChar));
    }
    
    return result;
  }
  
  /**
   * Read a list expression.
   * @param input Input stream.
   * @return Parsed expression.
   * @throws ParserException In case of a syntax or I/O error.
   */
  private SExpr parseList(final Reader input) {
    final List<SExpr> result = new LinkedList<SExpr>();
    read(input); // Skip left parenthesis.
    
    skipWhitespace(input);
    while (peek(input) != ')') {
      result.add(parse(input));
      skipWhitespace(input);
    }
    
    read(input); // Skip right parenthesis.
    
    return factory.createList(result);
  }
 
  /**
   * Read an unquoted symbol.
   * @param input Input stream.
   * @return Parsed expression.
   * @throws ParserException In case of a syntax or I/O error.
   */
  private SExpr parseSymbol(final Reader input) {
    final StringBuilder result = new StringBuilder();
    
    while (isSymbolPart(peek(input))) {
      result.appendCodePoint(read(input));
    }
    
    return factory.createSymbol(result.toString());
  }
  
  /**
   * Read a decimal (integer or rational).
   * @param input Input stream.
   * @return Parsed expression.
   * @throws ParserException In case of a syntax or I/O error.
   */
  private SExpr parseDec(final Reader input) {
    final StringBuilder result = new StringBuilder();
    
    while (Character.isDigit(peek(input))) {
      result.appendCodePoint(read(input));
    }
    if (peek(input) == '.') {
      result.appendCodePoint(read(input));
      while (Character.isDigit(peek(input))) {
        result.appendCodePoint(read(input));
      }
      final BigDecimal value;
      try {
        value = new BigDecimal(result.toString());
      } catch (final NumberFormatException e) {
        throw new ParserException("Invalid number", e);
      }
      return factory.createDecimal(value);      
    } else {
      final BigInteger value;
      try {
        value = new BigInteger(result.toString());
      } catch (final NumberFormatException e) {
        throw new ParserException("Invalid number", e);
      }
      return factory.createInteger(value);
    }
  }
  
  /**
   * Read a binary or hexadecimal numeral.
   * @param input Input stream.
   * @return Parsed expression.
   * @throws ParserException In case of a syntax or I/O error.
   */
  private SExpr parseHexOrBin(final Reader input) {
    final int kind;
    
    read(input); // Leading '#'
    
    kind = read(input);
    if (kind == 'x') {
      return parseHex(input);
    } else if (kind == 'b') {
      return parseBin(input);
    } else if (kind == -1) {
      throw new ParserException("End of stream inside numeral");
    } else {
      throw new ParserException(
          String.format("Unknown numeral type (%c)", kind));
    }
  }
  
  /**
   * Read a hexadecimal numeral.
   * @param input Input stream.
   * @return Parsed expression.
   * @throws ParserException In case of a syntax or I/O error.
   */
  private SExpr parseHex(final Reader input) {
    final String hexDigits = "0123456789abcdefABCDEF";
    final StringBuilder result = new StringBuilder();
    
    while (hexDigits.indexOf(peek(input)) != -1) {
      result.appendCodePoint(read(input));
    }
    
    final BigInteger value = new BigInteger(result.toString(), 16);
    return factory.createHex(value, result.length());
  }
  
  /**
   * Read a binary numeral.
   * @param input Input stream.
   * @return Parsed expression.
   * @throws ParserException In case of a syntax or I/O error.
   */
  private SExpr parseBin(final Reader input) {
    final StringBuilder result = new StringBuilder();
    
    int ch = peek(input);
    while (ch == '0' || ch == '1') {
      result.appendCodePoint(read(input));
      ch = peek(input);
    }
    
    final BigInteger value = new BigInteger(result.toString(), 2);
    return factory.createHex(value, result.length());
  }

  /**
   * Read a keyword.
   * @param input Input stream.
   * @return Parsed expression.
   * @throws ParserException In case of a syntax or I/O error.
   */
  private SExpr parseKeyword(final Reader input) {
    read(input); // Skip leading colon.
    final StringBuilder result = new StringBuilder();
    
    while (isSymbolPart(peek(input))) {
      result.appendCodePoint(read(input));
    }
    
    return factory.createKeyword(result.toString());    
  }
  
  /**
   * Read a quoted string.
   * @param input Input stream.
   * @return Parsed expression.
   * @throws ParserException In case of a syntax or I/O error.
   */
  private SExpr parseString(final Reader input) {
    read(input); // Skip first quote
    
    final StringBuilder result = new StringBuilder();
    
    int ch = read(input);
    while (ch != -1 && ch != '"') {
      if (ch == '\\') {
        final int escapedChar = read(input);
        if (escapedChar != '\\' && escapedChar != '"') {
          result.append('\\');
        }
        ch = escapedChar;
      } else {
        result.appendCodePoint(ch);
        ch = read(input);
      }
    }
    if (ch == -1) {
      throw new ParserException("End of stream inside quoted string");
    }
    
    return factory.createString(result.toString());    
  }

  /**
   * Read a quoted symbol.
   * @param input Input stream.
   * @return Parsed expression.
   * @throws ParserException In case of a syntax or I/O error.
   */
  private SExpr parseQuotedSymbol(final Reader input) {
    read(input); // Skip first quote
    
    final StringBuilder result = new StringBuilder();
    
    int ch = read(input);
    while (ch != -1 && ch != '|' && ch != '\\') {
      result.appendCodePoint(ch);
      ch = read(input);
    }
    if (ch == -1) {
      throw new ParserException("End of stream inside quoted symbol");
    } else if (ch == '\\') {
      throw new ParserException("Backslash inside quoted symbol.");
    }
    
    return factory.createSymbol(result.toString());    
  }
  
  /** 
   * Read a character and push it back to the stream.
   * @param input Input stream.
   * @return Character or -1 (end of stream).
   * @throws ParserException In case of an I/O error.
   */
  private int peek(final Reader input) {
    int result;
    
    try {
      input.mark(2);
      result = input.read();
      input.reset();
    } catch (final IOException e) {
      throw new ParserException("I/O error", e);
    }
    
    return result;
  }

  /** 
   * Read a single character.
   * @param input Input stream.
   * @return Character or -1 (end of stream).
   * @throws ParserException In case of an I/O error.
   */
  private int read(final Reader input) {
    int result;
    
    try {
      result = input.read();
    } catch (final IOException e) {
      throw new ParserException("I/O error", e);
    }
    
    return result;
  }
  
  /**
   * Discard leading whitespace from the input stream.
   * @param input Input stream.
   * @throws ParserException In case of an I/O error.
   */
  private void skipWhitespace(final Reader input) {
    int ch = peek(input);
    while (ch != -1 && Character.isWhitespace(ch)) {
      read(input);
      ch = peek(input);
    }
  }
  
  /**
   * Check if a character can start an unquoted symbol.
   * @param ch Character.
   * @return .
   */
  private boolean isSymbolStart(final int ch) {
    return Character.isLetter(ch) || 
           OPERATOR_CHARS.indexOf(ch) != -1;
  }

  /**
   * Check if a character can be a part of an unquoted symbol.
   * @param ch Character.
   * @return .
   */
  private boolean isSymbolPart(final int ch) {
    return Character.isLetterOrDigit(ch) || 
           OPERATOR_CHARS.indexOf(ch) != -1;
  }
  
  /** Class of exceptions thrown by the parser. */  
  public static class ParserException extends RuntimeException {
    /** SerialVersionUID */
    private static final long serialVersionUID = 1L;
    
    /** Constructor. */
    public ParserException() { }
    
    /**
     * Constructor.
     * @param msg Error message.
     */
    public ParserException(final String msg) { super(msg); }
    
    /**
     * Constructor.
     * @param cause Error cause.
     */
    public ParserException(final Throwable cause) { super(cause); }
    
    /**
     * Constructor.
     * @param msg Error message.
     * @param cause Error cause.
     */
    public ParserException(final String msg, 
                           final Throwable cause) { super(msg, cause); }
    
  }
}
