package net.afyre.haha.demo.solver.expr.impl;

import java.lang.ref.WeakReference;
import java.math.BigInteger;
import java.util.WeakHashMap;

import net.afyre.haha.demo.solver.expr.Expression;
import net.afyre.haha.demo.solver.expr.ExpressionFactory;
import net.afyre.haha.demo.solver.expr.ExpressionSwitch;
import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.formulae.FormulaSwitch;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.terms.TermSwitch;
import net.afyre.haha.demo.solver.expr.types.AtomicType;
import net.afyre.haha.demo.solver.expr.types.AtomicTypeSwitch;
import net.afyre.haha.demo.solver.expr.types.NonValueType;
import net.afyre.haha.demo.solver.expr.types.NonValueTypeSwitch;
import net.afyre.haha.demo.solver.expr.types.StructuredType;
import net.afyre.haha.demo.solver.expr.types.StructuredTypeSwitch;
import net.afyre.haha.demo.solver.expr.types.TypeSwitch;
import net.afyre.haha.demo.solver.expr.types.ValueType;
import net.afyre.haha.demo.solver.expr.types.ValueTypeSwitch;

/**
 * A canonicalizing wrapper for an expression factory.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 */
public final class CanonicalizingExpressionFactory 
    implements ExpressionFactory {

  /** Factory used to create instances. */
  private final ExpressionFactory factory;

  /** Hashmap (common for all types). */
  private final WeakHashMap<Expression, WeakReference<Expression>> instances;

  /**
   * Constructor.
   * @param factory Factory used to create instances.
   */
  private CanonicalizingExpressionFactory(final ExpressionFactory factory) {
    this.factory = factory;
    instances = new WeakHashMap<Expression, WeakReference<Expression>>();
  }

  /**
   * A wrapper for expressions which uses reference equality for comparisons.
   *
   * @param <T> Type of wrapped exceptions.
   */
  private static class CanonicalExpression<T extends Expression> 
      extends AbstractExpression {
    private final T expression;
    
    public CanonicalExpression(final T expression) {
      this.expression = expression;
    }

    /** {@inheritDoc} */
    @Override
    public boolean equals(final Object obj) {
      if (obj instanceof CanonicalExpression) {
        return this == obj;
      } else {
        return expression.equals(obj);
      }
    }

    /** {@inheritDoc} */
    @Override
    public int hashCode() {
      return expression.hashCode();
    }
    
    /** {@inheritDoc} */
    @Override
    public String toString() {
      return expression.toString();
    }
    
    /** {@inheritDoc} */
    @Override
    public <Result> Result accept(final ExpressionSwitch<? extends Result> v) {
      return expression.accept(v);
    }
    
    protected final T getWrappedExpression() {
      return expression;
    }
  }
  
  /**
   * A {@link CanonicalExpression} wrapper for formulae. 
   */
  private static final class CanonicalFormula 
      extends CanonicalExpression<Formula> implements Formula {
    
    public CanonicalFormula(final Formula formula) {
      super(formula);
    }
    
    /** {@inheritDoc} */
    @Override
    public <Result> Result accept(final FormulaSwitch<? extends Result> v) {
      return getWrappedExpression().accept(v);
    }
  }

  /**
   * A {@link CanonicalExpression} wrapper for terms. 
   */
  private static final class CanonicalTerm 
      extends CanonicalExpression<Term> implements Term {
    
    public CanonicalTerm(final Term term) {
      super(term);
    }
    
    /** {@inheritDoc} */
    @Override
    public <Result> Result accept(final TermSwitch<? extends Result> v) {
      return getWrappedExpression().accept(v);
    }
  }

  /**
   * A {@link CanonicalExpression} wrapper for structured types. 
   */  
  private static final class CanonicalStructuredType 
      extends CanonicalExpression<StructuredType> implements StructuredType {
    
    public CanonicalStructuredType(final StructuredType type) {
      super(type);
    }

    /** {@inheritDoc} */
    @Override
    public <Result> Result accept(final ValueTypeSwitch<? extends Result> v) {
      return getWrappedExpression().accept(v);
    }

    /** {@inheritDoc} */
    @Override
    public <Result> Result accept(final TypeSwitch<? extends Result> v) {
      return getWrappedExpression().accept(v);
    }

    /** {@inheritDoc} */
    @Override
    public <Result> Result accept(
        final StructuredTypeSwitch<? extends Result> v) {
      return getWrappedExpression().accept(v);
    }
  }

  /**
   * A {@link CanonicalExpression} wrapper for non-value types. 
   */
  private static final class CanonicalNonValueType 
      extends CanonicalExpression<NonValueType> implements NonValueType {

    public CanonicalNonValueType(final NonValueType type) {
      super(type);
    }

    /** {@inheritDoc} */
    @Override
    public <Result> Result accept(final TypeSwitch<? extends Result> v) {
      return getWrappedExpression().accept(v);
    }

    /** {@inheritDoc} */
    @Override
    public <Result> Result accept(
        final NonValueTypeSwitch<? extends Result> v) {
      return getWrappedExpression().accept(v);
    }
  }  

  /**
   * A {@link CanonicalExpression} wrapper for atomic types. 
   */
  private static final class CanonicalAtomicType 
    extends CanonicalExpression<AtomicType> implements AtomicType {

    public CanonicalAtomicType(final AtomicType type) {
      super(type);
    }
  
    /** {@inheritDoc} */
    @Override
    public <Result> Result accept(final ValueTypeSwitch<? extends Result> v) {
      return getWrappedExpression().accept(v);
    }
  
    /** {@inheritDoc} */
    @Override
    public <Result> Result accept(final TypeSwitch<? extends Result> v) {
      return getWrappedExpression().accept(v);
    }
  
    /** {@inheritDoc} */
    @Override
    public <Result> Result accept(
        final AtomicTypeSwitch<? extends Result> v) {
      return getWrappedExpression().accept(v);
    }
  } 
  
  /**
   * Create an instance of this factory class.
   * @param factory Factory used to create instances.
   * @return Instance.
   */
  public static CanonicalizingExpressionFactory 
  getInstance(final ExpressionFactory factory) {
    return new CanonicalizingExpressionFactory(factory);
  }

  /**
   * Get the canonical instance for given expression.
   * @param candidate Expression to canonicalize.
   * @return Canonicalized expression.
   */
  private synchronized Expression getCanonicalInstance(
      final CanonicalExpression<?> candidate) {
    if (candidate == null || candidate.getWrappedExpression() == null) {
      return null;
    }
    final Expression expr = candidate.getWrappedExpression();
    final WeakReference<Expression> ref = instances.get(expr);
    Expression result = null;
    if (ref != null) { result = ref.get(); }
    if (result == null) {
      instances.put(candidate, new WeakReference<Expression>(candidate));
      result = candidate;
    }
    return result;
  }

  /**
   * Get the canonical instance for given formula.
   * @param formula Formula to canonicalize.
   * @return Canonicalized formula.
   */
  private Formula getCanonicalFormula(final Formula formula) {
    return (Formula) getCanonicalInstance(new CanonicalFormula(formula));
  }

  /**
   * Get the canonical instance for given term.
   * @param term Term to canonicalize.
   * @return Canonicalized term.
   */
  private Term getCanonicalTerm(final Term term) {
    return (Term) getCanonicalInstance(new CanonicalTerm(term));
  }
  
  /**
   * Get the canonical instance for given structure type.
   * @param type Type to canonicalize.
   * @return Canonicalized type.
   */
  private StructuredType getCanonicalStructuredType(
                             final StructuredType type) {
    return (StructuredType) getCanonicalInstance(
                                new CanonicalStructuredType(type));
  }
  
  /**
   * Get the canonical instance for given non-value type.
   * @param type Type to canonicalize.
   * @return Canonicalized type.
   */
  private NonValueType getCanonicalNonValueType(final NonValueType type) {
    return (NonValueType) getCanonicalInstance(
                               new CanonicalNonValueType(type));
  }
  
  /**
   * Get the canonical instance for given atomic type.
   * @param type Type to canonicalize.
   * @return Canonicalized type.
   */
  private AtomicType getCanonicalAtomicType(final AtomicType type) {
    return (AtomicType) getCanonicalInstance(
                               new CanonicalAtomicType(type));
  }
  
  /** {@inheritDoc} */
  @Override
  public Term createIntConst(final BigInteger value) {
    return getCanonicalTerm(factory.createIntConst(value));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVariable(final String var) {
    return getCanonicalTerm(factory.createVariable(var));
  }

  /** {@inheritDoc} */
  @Override
  public Term createMkTuple(final Term... elements) {
    return getCanonicalTerm(factory.createMkTuple(elements));
  }

  /** {@inheritDoc} */
  @Override
  public Term createTupleAccess(final Term tuple, final int index) {
    return getCanonicalTerm(factory.createTupleAccess(tuple, index));
  }

  /** {@inheritDoc} */
  @Override
  public Term createUminus(final Term arg) {
    return getCanonicalTerm(factory.createUminus(arg));
  }

  /** {@inheritDoc} */
  @Override
  public Term createAddition(final Term left, final Term right) {
    return getCanonicalTerm(factory.createAddition(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createSubtraction(final Term left, final Term right) {
    return getCanonicalTerm(factory.createSubtraction(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createMultiplication(final Term left, final Term right) {
    return getCanonicalTerm(
                      factory.createMultiplication(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createDivision(final Term left, final Term right) {
    return getCanonicalTerm(factory.createDivision(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createRemainder(final Term left, final Term right) {
    return getCanonicalTerm(factory.createRemainder(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createPower(final Term left, final Term right) {
    return getCanonicalTerm(factory.createPower(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createLet(final String var, final ValueType type, 
                        final Term value, final Term body) {
    return getCanonicalTerm(
                      factory.createLet(var, type, value, body));
  }

  /** {@inheritDoc} */
  @Override
  public Term createLetProp(final String var, 
                            final Formula value, 
                            final Term body) {
    return getCanonicalTerm(factory.createLetProp(var, value, body));
  }

  /** {@inheritDoc} */
  @Override
  public Term createIfThenElse(final Formula cond, 
                               final Term ifTrue,
                               final Term ifFalse) {
    return getCanonicalTerm(
                      factory.createIfThenElse(cond, ifTrue, ifFalse));
  }

  /** {@inheritDoc} */
  @Override
  public Term createArrayAccess(final Term arr, final Term index) {
    return getCanonicalTerm(factory.createArrayAccess(arr, index));
  }

  /** {@inheritDoc} */
  @Override
  public Term createArrayUpdate(final Term arr, 
                                final Term index, 
                                final Term value) {
    return getCanonicalTerm(
                      factory.createArrayUpdate(arr, index, value));
  }

  /** {@inheritDoc} */
  @Override
  public Term createCall(final String func, final Term... args) {
    return getCanonicalTerm(factory.createCall(func, args));
  }

  /** {@inheritDoc} */
  @Override
  public Term createConcat(final Term left, final Term right) {
    return getCanonicalTerm(factory.createConcat(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createSelectBits(final Term vec, 
                               final int left, 
                               final int right) {
    return getCanonicalTerm(
                      factory.createSelectBits(vec, left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createShl(final Term vec, final int offset) {
    return getCanonicalTerm(factory.createShl(vec, offset));
  }

  /** {@inheritDoc} */
  @Override
  public Term createShr(final Term vec, final int offset) {
    return getCanonicalTerm(factory.createShr(vec, offset));
  }

  /** {@inheritDoc} */
  @Override
  public Term createSX(final Term vec, final int offset) {
    return getCanonicalTerm(factory.createSX(vec, offset));
  }

  /** {@inheritDoc} */
  @Override
  public Term createZeroExtend(final Term vec, final int offset) {
    return getCanonicalTerm(factory.createZeroExtend(vec, offset));
  }

  /** {@inheritDoc} */
  @Override
  public Term createRepeat(final Term vec, final int count) {
    return getCanonicalTerm(factory.createRepeat(vec, count));
  }

  /** {@inheritDoc} */
  @Override
  public Term createRotateLeft(final Term vec, final int offset) {
    return getCanonicalTerm(factory.createRotateLeft(vec, offset));
  }

  /** {@inheritDoc} */
  @Override
  public Term createRotateRight(final Term vec, final int offset) {
    return getCanonicalTerm(factory.createRotateRight(vec, offset));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorAnd(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorAnd(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorOr(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorOr(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorXor(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorXor(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorNand(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorNand(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorNor(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorNor(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorXnor(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorXnor(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorCmp(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorCmp(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorSum(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorSum(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorDiff(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorDiff(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorUminus(final Term vec) {
    return getCanonicalTerm(factory.createVectorUminus(vec));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorProd(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorProd(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorShl(final Term vec, final Term offset) {
    return getCanonicalTerm(factory.createVectorShl(vec, offset));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorLshr(final Term vec, final Term offset) {
    return getCanonicalTerm(factory.createVectorLshr(vec, offset));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorRshr(final Term vec, final Term offset) {
    return getCanonicalTerm(factory.createVectorRshr(vec, offset));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorUdiv(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorUdiv(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorSdiv(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorSdiv(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorUrem(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorUrem(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorSrem(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorSrem(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorSmod(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorSmod(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorLt(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorLt(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorGt(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorGt(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorLe(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorLe(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorGe(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorGe(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorSlt(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorSlt(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorSgt(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorSgt(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorSle(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorSle(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Term createVectorSge(final Term left, final Term right) {
    return getCanonicalTerm(factory.createVectorSge(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createTrue() {
    return getCanonicalFormula(factory.createTrue());
  }

  /** {@inheritDoc} */
  @Override
  public Formula createFalse() {
    return getCanonicalFormula(factory.createFalse());
  }

  /** {@inheritDoc} */
  @Override
  public Formula createFormulaVariable(final String name) {
    return getCanonicalFormula(factory.createFormulaVariable(name));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createNegation(final Formula arg) {
    return getCanonicalFormula(factory.createNegation(arg));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createConjunction(final Formula left, final Formula right) {
    return getCanonicalFormula(
                         factory.createConjunction(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createDisjunction(final Formula left, final Formula right) {
    return getCanonicalFormula(
                         factory.createDisjunction(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createImplication(final Formula left, final Formula right) {
    return getCanonicalFormula(
                         factory.createImplication(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createXor(final Formula left, final Formula right) {
    return getCanonicalFormula(factory.createXor(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createEquivalence(final Formula left, final Formula right) {
    return getCanonicalFormula(
                         factory.createEquivalence(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createEq(final Term left, final Term right) {
    return getCanonicalFormula(factory.createEq(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createNeq(final Term left, final Term right) {
    return getCanonicalFormula(factory.createNeq(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createDistinct(final Term... args) {
    return getCanonicalFormula(factory.createDistinct(args));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createLt(final Term left, final Term right) {
    return getCanonicalFormula(factory.createLt(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createGt(final Term left, final Term right) {
    return getCanonicalFormula(factory.createGt(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createLe(final Term left, final Term right) {
    return getCanonicalFormula(factory.createLe(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createGe(final Term left, final Term right) {
    return getCanonicalFormula(factory.createGe(left, right));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createExists(final Iterable<VariableBinding> variables, 
                              final Formula arg,
                              final Expression... patterns) {
    return getCanonicalFormula(
                         factory.createExists(variables, arg, patterns));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createForall(final Iterable<VariableBinding> variables, 
                              final Formula arg,
                              final Expression... patterns) {
    return getCanonicalFormula(
                         factory.createForall(variables, arg, patterns));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createFLet(final String var, 
                            final ValueType type, 
                            final Term value,
                            final Formula body) {
    return getCanonicalFormula(
                         factory.createFLet(var, type, value, body));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createFLetProp(final String var, 
                                final Formula value, 
                                final Formula body) {
    return getCanonicalFormula(
                         factory.createFLetProp(var, value, body));
  }

  /** {@inheritDoc} */
  @Override
  public Formula createPredicateCall(final String pred, final Term... args) {
    return getCanonicalFormula(
                         factory.createPredicateCall(pred, args));
  }

  /** {@inheritDoc} */
  @Override
  public AtomicType createRational() {
    return getCanonicalAtomicType(factory.createRational());
  }

  /** {@inheritDoc} */
  @Override
  public AtomicType createInt() {
    return getCanonicalAtomicType(factory.createInt());
  }

  /** {@inheritDoc} */
  @Override
  public AtomicType createBitvector(final int size) {
    return getCanonicalAtomicType(factory.createBitvector(size));
  }

  /** {@inheritDoc} */
  @Override
  public StructuredType createArray(final ValueType indexType, 
                                    final ValueType valueType) {
    return getCanonicalStructuredType(
                                factory.createArray(indexType, valueType));
  }

  /** {@inheritDoc} */
  @Override
  public StructuredType createTuple(final ValueType... types) {
    return getCanonicalStructuredType(factory.createTuple(types));
  }

  /** {@inheritDoc} */
  @Override
  public NonValueType createProp() {
    return getCanonicalNonValueType(factory.createProp());
  }

  /** {@inheritDoc} */
  @Override
  public NonValueType createFunction(final ValueType resultType, 
                                     final ValueType... argTypes) {
    return getCanonicalNonValueType(
                              factory.createFunction(resultType, argTypes));
  }

  /** {@inheritDoc} */
  @Override
  public NonValueType createPredicate(final ValueType... argTypes) {
    return getCanonicalNonValueType(
                              factory.createPredicate(argTypes));
  }
}
