package net.afyre.haha.demo.solver.expr.util;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import net.afyre.haha.demo.solver.expr.Expression;
import net.afyre.haha.demo.solver.expr.ExpressionSwitch;
import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.types.ValueType;

/**
 * Switch used to convert expressions into arrays consisting 
 * of the constructor name and arguments.
 * Arguments which are expressions are *not* unpacked recursively.
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public final class ExprToArraySwitch 
    implements ExpressionSwitch<Object[]> {
  /** Singleton instance. */
  private static final ExprToArraySwitch INSTANCE = new ExprToArraySwitch();
   
  /**
   * Get an instance of this class.
   * @return Instance.
   */
  public static ExprToArraySwitch getInstance() {
    return INSTANCE;
  }
  
  /** {@inheritDoc} */
  @Override
  public Object[] caseIntConst(final BigInteger value) {
    final List<Object> result = new ArrayList<Object>();
    result.add("IntConst");
    result.add(value);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseFormulaVariable(final String var) {
    final List<Object> result = new ArrayList<Object>();
    result.add("FormulaVariable");
    result.add(var);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVariable(final String var) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Variable");
    result.add(var);
    return result.toArray();
  }
  
  /** {@inheritDoc} */
  @Override
  public Object[] caseMkTuple(final Term... elements) {
    final List<Object> result = new ArrayList<Object>();
    result.add("MkTuple");
    result.addAll(Arrays.asList(elements));
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseTupleAccess(final Term tuple, final int index) {
    final List<Object> result = new ArrayList<Object>();
    result.add("TupleAccess");
    result.add(tuple);
    result.add(index);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseUminus(final Term arg) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Uminus");
    result.add(arg);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseAddition(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Addition");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseSubtraction(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Subtraction");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseMultiplication(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Multiplication");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseDivision(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Division");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseRemainder(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Remainder");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] casePower(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Power");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseLet(final String var,
                          final ValueType type,
                          final Term value,
                          final Term body) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Let");
    result.add(var);
    result.add(type);
    result.add(value);
    result.add(body);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseLetProp(final String var,
                              final Formula value,
                              final Term body) {
    final List<Object> result = new ArrayList<Object>();
    result.add("LetProp");
    result.add(var);
    result.add(value);
    result.add(body);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseIfThenElse(final Formula cond,
                                 final Term ifTrue,
                                 final Term ifFalse) {
    final List<Object> result = new ArrayList<Object>();
    result.add("IfThenElse");
    result.add(cond);
    result.add(ifTrue);
    result.add(ifFalse);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseArrayAccess(final Term arr, final Term index) {
    final List<Object> result = new ArrayList<Object>();
    result.add("ArrayAccess");
    result.add(arr);
    result.add(index);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseArrayUpdate(final Term arr,
                                  final Term index,
                                  final Term value) {
    final List<Object> result = new ArrayList<Object>();
    result.add("ArrayUpdate");
    result.add(arr);
    result.add(index);
    result.add(value);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseCall(final String func, final Term... args) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Call");
    result.add(func);
    result.addAll(Arrays.asList(args));
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseConcat(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Concat");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseSelectBits(final Term vec,
                                 final int left,
                                 final int right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("SelectBits");
    result.add(vec);
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseShl(final Term vec, final int offset) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Shl");
    result.add(vec);
    result.add(offset);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseShr(final Term vec, final int offset) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Shr");
    result.add(vec);
    result.add(offset);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseSX(final Term vec, final int offset) {
    final List<Object> result = new ArrayList<Object>();
    result.add("SX");
    result.add(vec);
    result.add(offset);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseZeroExtend(final Term vec, final int offset) {
    final List<Object> result = new ArrayList<Object>();
    result.add("ZeroExtend");
    result.add(vec);
    result.add(offset);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseRepeat(final Term vec, final int count) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Repeat");
    result.add(vec);
    result.add(count);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseRotateLeft(final Term vec, final int offset) {
    final List<Object> result = new ArrayList<Object>();
    result.add("RotateLeft");
    result.add(vec);
    result.add(offset);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseRotateRight(final Term vec, final int offset) {
    final List<Object> result = new ArrayList<Object>();
    result.add("RotateRight");
    result.add(vec);
    result.add(offset);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorAnd(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorAnd");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorOr(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorOr");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorXor(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorXor");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorNand(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorNand");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorNor(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorNor");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorXnor(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorXnor");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorCmp(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorCmp");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorSum(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorSum");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorDiff(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorDiff");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorUminus(final Term vec) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorUminus");
    result.add(vec);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorProd(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorProd");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorShl(final Term vec, final Term offset) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorShl");
    result.add(vec);
    result.add(offset);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorLshr(final Term vec, final Term offset) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorLshr");
    result.add(vec);
    result.add(offset);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorRshr(final Term vec, final Term offset) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorRshr");
    result.add(vec);
    result.add(offset);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorUdiv(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorUdiv");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorSdiv(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorSdiv");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorUrem(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorUrem");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorSrem(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorSrem");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorSmod(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorSmod");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorLt(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorLt");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorGt(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorGt");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorLe(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorLe");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorGe(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorGe");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorSlt(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorSlt");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorSgt(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorSgt");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorSle(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorSle");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseVectorSge(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("VectorSge");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseTrue() {
    final List<Object> result = new ArrayList<Object>();
    result.add("True");
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseFalse() {
    final List<Object> result = new ArrayList<Object>();
    result.add("False");
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseNegation(final Formula arg) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Negation");
    result.add(arg);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseConjunction(final Formula left,
                                  final Formula right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Conjunction");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseDisjunction(final Formula left,
                                  final Formula right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Disjunction");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseImplication(final Formula left,
                                  final Formula right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Implication");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseXor(final Formula left, final Formula right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Xor");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseEquivalence(final Formula left,
                                  final Formula right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Equivalence");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseEq(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Eq");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseNeq(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Neq");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseDistinct(final Term... args) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Distinct");
    result.addAll(Arrays.asList(args));
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseLt(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Lt");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseGt(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Gt");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseLe(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Le");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseGe(final Term left, final Term right) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Ge");
    result.add(left);
    result.add(right);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseExists(final Iterable<VariableBinding> variables,
                             final Formula arg,
                             final Expression... patterns) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Exists");
    for (final VariableBinding variable : variables) {
      result.add(variable);
    }
    result.add(arg);
    // NOTE: Instantiation patterns are ignored.
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseForall(final Iterable<VariableBinding> variables,
                             final Formula arg,
                             final Expression... patterns) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Forall");
    for (final VariableBinding variable : variables) {
      result.add(variable);
    }
    result.add(arg);
    // NOTE: Instantiation patterns are ignored.
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseFLet(final String var,
                          final ValueType type,
                          final Term value,
                          final Formula body) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Let");
    result.add(var);
    result.add(type);
    result.add(value);
    result.add(body);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseFLetProp(final String var,
                              final Formula value,
                              final Formula body) {
    final List<Object> result = new ArrayList<Object>();
    result.add("LetProp");
    result.add(var);
    result.add(value);
    result.add(body);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseRational() {
    final List<Object> result = new ArrayList<Object>();
    result.add("Rational");
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseInt() {
    final List<Object> result = new ArrayList<Object>();
    result.add("Int");
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseBitvector(final int size) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Bitvector");
    result.add(size);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseArray(final ValueType indexType,
                            final ValueType valueType) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Array");
    result.add(indexType);
    result.add(valueType);
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseTuple(final ValueType... types) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Tuple");
    result.addAll(Arrays.asList(types));
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseProp() {
    final List<Object> result = new ArrayList<Object>();
    result.add("Prop");
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] caseFunction(final ValueType resultType,
                               final ValueType... argTypes) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Function");
    result.add(resultType);
    result.addAll(Arrays.asList(argTypes));
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] casePredicate(final ValueType... argTypes) {
    final List<Object> result = new ArrayList<Object>();
    result.add("Predicate");
    result.addAll(Arrays.asList(argTypes));
    return result.toArray();
  }

  /** {@inheritDoc} */
  @Override
  public Object[] casePredicateCall(final String pred, final Term... args) {
    final List<Object> result = new ArrayList<Object>();
    result.add("PredicateCall");
    result.add(pred);
    result.addAll(Arrays.asList(args));
    return result.toArray();
  }
}
