package net.afyre.haha.demo.util.tests;

import static org.junit.Assert.*;
import static org.junit.Assume.*;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import net.afyre.haha.demo.util.CmpUtil;

import org.junit.Test;
import org.junit.experimental.runners.Enclosed;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

/**
 * Tests for the {@link CmpUtil} class.
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
@RunWith(Enclosed.class)
public class CmpUtilTest {
  
  /** Tests with no parameters. */
  public static class Tests {
    /** GetInstance does not return null. */
    @Test
    public void testInstanceNotNull() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assertNotNull("Null returned from getInstance()", cmpUtil);
    }
    
    /** SafeEquals returns true for two null arguments. */
    @Test
    public void testEqualsNullAndNull() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertTrue("Wrong comparison result", 
                 cmpUtil.safeEquals(null, null));
    }
  
    /** SafeEquals returns false when only the first argument is null. */
    @Test
    public void testEqualsNotNullAndNull() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertFalse("Wrong comparison result",  
                  cmpUtil.safeEquals(42, null));
    }
    
    /** SafeEquals returns false when only the second argument is null. */
    @Test
    public void testEqualsNullAndNotNull() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertFalse("Wrong comparison result",  
                  cmpUtil.safeEquals(null, 42));
    }
    
    /** SafeEquals returns false for two different non-null arguments. */
    @Test
    public void testEqualsWithDiffrentArgs() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertFalse("Wrong comparison result",  
                  cmpUtil.safeEquals(42, 44));    
    }
    
    /** SafeEquals returns true for equal, non-null arguments. */
    @Test
    public void testEqualsTrue() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertTrue("Wrong comparison result",  
                 cmpUtil.safeEquals(42, 42));        
    }
  
    /** SafeEquals returns true for two null arrays. */
    @Test
    public void testArrayEqualsNullAndNull() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertTrue("Wrong comparison result", 
                 cmpUtil.safeEquals((Object[])null, null));
    }
  
    /** SafeEquals returns false when only the first array is null. */
    @Test
    public void testArrayEqualsNotNullAndNull() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertFalse("Wrong comparison result",  
                  cmpUtil.safeEquals(new Integer[]{}, null));
    }
    
    /** SafeEquals returns false when only the second array is null. */
    @Test
    public void testArrayEqualsNullAndNotNull() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertFalse("Wrong comparison result",  
                  cmpUtil.safeEquals(null, new Integer[]{}));
    }
    
    /** SafeEquals returns false for two different non-null arrays. */
    @Test
    public void testArrayEqualsWithDiffrentArgs() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertFalse("Wrong comparison result",  
                  cmpUtil.safeEquals(new Integer[]{42}, 
                                     new Integer[]{44}));    
    }
    
    /** SafeEquals returns true for equal, non-null arrays. */
    @Test
    public void testArrayEqualsTrue() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertTrue("Wrong comparison result",  
                 cmpUtil.safeEquals(new Integer[]{42}, 
                                    new Integer[]{42}));        
    }
  
    /** SafeEquals returns false for arrays of different size. */
    @Test
    public void testArrayEqualsWithDiffrentSizes() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertFalse("Wrong comparison result",  
                  cmpUtil.safeEquals(new Integer[]{1, 1}, 
                                     new Integer[]{1, 1, 1}));    
    }
  
    /** IterEquals returns true for two null iterables. */
    @Test
    public void testIterEqualsNullAndNull() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      final Iterable<Integer> it1 = null;
      final Iterable<Integer> it2 = null;
      assertTrue("Wrong comparison result", 
                 cmpUtil.iterEquals(it1, it2));
    }
  
    /** IterEquals returns false when only the first iterable is null. */
    @Test
    public void testIterEqualsNotNullAndNull() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      final Iterable<Integer> it1 = 
          Arrays.asList(new Integer[] { 42 });
      final Iterable<Integer> it2 = null;
      assertFalse("Wrong comparison result",  
                  cmpUtil.iterEquals(it1, it2));
    }
    
    /** 
     * IterEquals returns false when only the second iterable is null.
     */
    @Test
    public void testIterEqualsNullAndNotNull() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      final Iterable<Integer> it1 = null;
      final Iterable<Integer> it2 = 
          Arrays.asList(new Integer[] { 42 });
      assertFalse("Wrong comparison result",  
                  cmpUtil.iterEquals(it1, it2));
    }
    
    /** IterEquals returns false for two different non-null iterables. */
    @Test
    public void testIterEqualsWithDiffrentArgs() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      final Iterable<Integer> it1 = 
          Arrays.asList(new Integer[] { 42, 42 });
      final Iterable<Integer> it2 = 
          Arrays.asList(new Integer[] { 42, 44 });
      assertFalse("Wrong comparison result",  
                  cmpUtil.iterEquals(it1, it2));  
    }
    
    /** IterEquals returns true for equal, non-null iterables. */
    @Test
    public void testIterEqualsTrue() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      final Iterable<Integer> it1 = 
          Arrays.asList(new Integer[] { 42, 42 });
      final Iterable<Integer> it2 = 
          Arrays.asList(new Integer[] { 42, 42 });
      assertTrue("Wrong comparison result",  
                 cmpUtil.iterEquals(it1, it2));      
    }
  
    /** IterEquals returns false for iterables of different size. */
    @Test
    public void testIterEqualsWithDiffrentSizes() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      final Iterable<Integer> it1 = 
          Arrays.asList(new Integer[] { 42, 42 });
      final Iterable<Integer> it2 = 
          Arrays.asList(new Integer[] { 42, 42, 42 });
      assertFalse("Wrong comparison result",  
                  cmpUtil.iterEquals(it1, it2)); 
    }
  
  }
  /** Test compare() on single (possibly null) objects. */
  @RunWith(Parameterized.class)
  public static class CompareTest {
    /** First object. */
    private final Integer first;
    
    /** Second object. */
    private final Integer second;
    
    /** Expected comparison result. */
    private final int expected;
    
    /**
     * Constructor.
     * @param first First object to be compared.
     * @param second Second object to be compared.
     * @param expected Expected comparison result.
     */
    public CompareTest(final Integer first, 
                       final Integer second, 
                       final int expected) {
      this.first = first;
      this.second = second;
      this.expected = expected;
    }
    
    @Parameters
    public static Collection<Object[]> data() {
      final Object[][] data = new Object[][] {
        { null, null, 0 }, 
        { null, 1, -1 }, 
        { null, 2, -1 }, 
        { 1, null, 1 }, 
        { 1, 1, 0 }, 
        { 1, 2, -1 }, 
        { 2, null, 1 }, 
        { 2, 1, 1 }, 
        { 2, 2, 0 }                                          
      };
      return Arrays.asList(data);
    }
    
    /** Comparison result is correct. */
    @Test
    public void testCompare() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertEquals(
         String.format("Wrong comparison result for data: %s, %s", 
                       first, second),
         expected,
         cmpUtil.compare(first, second));
    }
  }
  
  /** Test compare() on pairs of (possibly null) elements. */
  @RunWith(Parameterized.class)
  public static class ComparePairsTest {
    /** First component of the first pair. */
    private final Integer a1;
    
    /** Second component of the first pair. */
    private final Integer a2;
    
    /** First component of the second pair. */
    private final Integer b1;
    
    /** Second component of the second pair. */
    private final Integer b2;
    
    /** Expected comparison result. */
    private final int expected;
    
    /**
     * Constructor.
     * @param a1 First component of the first pair.
     * @param a2 Second component of the first pair.
     * @param b1 First component of the second pair.
     * @param b2 Second component of the second pair.
     * @param expected Expected comparison result.
     */
    public ComparePairsTest(final Integer a1,
                            final Integer a2,
                            final Integer b1,
                            final Integer b2,
                            final int expected) {
      this.a1 = a1;
      this.a2 = a2;
      this.b1 = b1;
      this.b2 = b2;
      this.expected = expected;
    }
    
    @Parameters
    public static Collection<Object[]> data() {
      final Object[][] data = new Object[][] {
        { null, null, null, null, 0 }, 
        { null, null, null, 1, -1 }, 
        { null, null, null, 2, -1 }, 
        { null, null, 1, null, -1 }, 
        { null, null, 1, 1, -1 }, 
        { null, null, 1, 2, -1 }, 
        { null, null, 2, null, -1 }, 
        { null, null, 2, 1, -1 }, 
        { null, null, 2, 2, -1 }, 
        { null, 1, null, null, 1 }, 
        { null, 1, null, 1, 0 }, 
        { null, 1, null, 2, -1 }, 
        { null, 1, 1, null, -1 }, 
        { null, 1, 1, 1, -1 }, 
        { null, 1, 1, 2, -1 }, 
        { null, 1, 2, null, -1 }, 
        { null, 1, 2, 1, -1 }, 
        { null, 1, 2, 2, -1 }, 
        { null, 2, null, null, 1 }, 
        { null, 2, null, 1, 1 }, 
        { null, 2, null, 2, 0 }, 
        { null, 2, 1, null, -1 }, 
        { null, 2, 1, 1, -1 }, 
        { null, 2, 1, 2, -1 }, 
        { null, 2, 2, null, -1 }, 
        { null, 2, 2, 1, -1 }, 
        { null, 2, 2, 2, -1 }, 
        { 1, null, null, null, 1 }, 
        { 1, null, null, 1, 1 }, 
        { 1, null, null, 2, 1 }, 
        { 1, null, 1, null, 0 }, 
        { 1, null, 1, 1, -1 }, 
        { 1, null, 1, 2, -1 }, 
        { 1, null, 2, null, -1 }, 
        { 1, null, 2, 1, -1 }, 
        { 1, null, 2, 2, -1 }, 
        { 1, 1, null, null, 1 }, 
        { 1, 1, null, 1, 1 }, 
        { 1, 1, null, 2, 1 }, 
        { 1, 1, 1, null, 1 }, 
        { 1, 1, 1, 1, 0 }, 
        { 1, 1, 1, 2, -1 }, 
        { 1, 1, 2, null, -1 }, 
        { 1, 1, 2, 1, -1 }, 
        { 1, 1, 2, 2, -1 }, 
        { 1, 2, null, null, 1 }, 
        { 1, 2, null, 1, 1 }, 
        { 1, 2, null, 2, 1 }, 
        { 1, 2, 1, null, 1 }, 
        { 1, 2, 1, 1, 1 }, 
        { 1, 2, 1, 2, 0 }, 
        { 1, 2, 2, null, -1 }, 
        { 1, 2, 2, 1, -1 }, 
        { 1, 2, 2, 2, -1 }, 
        { 2, null, null, null, 1 }, 
        { 2, null, null, 1, 1 }, 
        { 2, null, null, 2, 1 }, 
        { 2, null, 1, null, 1 }, 
        { 2, null, 1, 1, 1 }, 
        { 2, null, 1, 2, 1 }, 
        { 2, null, 2, null, 0 }, 
        { 2, null, 2, 1, -1 }, 
        { 2, null, 2, 2, -1 }, 
        { 2, 1, null, null, 1 }, 
        { 2, 1, null, 1, 1 }, 
        { 2, 1, null, 2, 1 }, 
        { 2, 1, 1, null, 1 }, 
        { 2, 1, 1, 1, 1 }, 
        { 2, 1, 1, 2, 1 }, 
        { 2, 1, 2, null, 1 }, 
        { 2, 1, 2, 1, 0 }, 
        { 2, 1, 2, 2, -1 }, 
        { 2, 2, null, null, 1 }, 
        { 2, 2, null, 1, 1 }, 
        { 2, 2, null, 2, 1 }, 
        { 2, 2, 1, null, 1 }, 
        { 2, 2, 1, 1, 1 }, 
        { 2, 2, 1, 2, 1 }, 
        { 2, 2, 2, null, 1 }, 
        { 2, 2, 2, 1, 1 }, 
        { 2, 2, 2, 2, 0 }                                        
      };
      return Arrays.asList(data);
    }
    
    /** Comparison result is correct. */
    @Test
    public void testCompare() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertEquals(
         String.format(
            "Wrong comparison result for data: (%s, %s), (%s, %s)", 
            a1, a2, b1, b2),
         expected,
         cmpUtil.compare(a1, a2, b1, b2));
    }
  }
  
  /**
   * Convert an array of integers and nulls to 
   * a string in a way which preserves ordering.
   * @param tuple Array of integers and nulls.
   * @return ...
   */
  private static String tupleToString(final Object[] tuple) {
    final StringBuilder builder = new StringBuilder();
    for (final Object obj : tuple) {
      final String element;
      if (obj == null) {
        element = "AnullA";
      } else {
        element = String.format("B%dB", obj);
      }
      builder.append(element);
    }
    return builder.toString();
  }
  
  /** Test compare() on triples of (possibly null) elements. */
  @RunWith(Parameterized.class)
  public static class CompareTriplesTest {
    /** First component of the first triple. */
    private final Integer a1;
    
    /** Second component of the first triple. */
    private final Integer a2;
    
    /** Third component of the first triple. */
    private final Integer a3;
    
    /** First component of the second triple. */
    private final Integer b1;
    
    /** Second component of the second triple. */
    private final Integer b2;

    /** Third component of the second triple. */
    private final Integer b3;
    
    /** Expected comparison result. */
    private final int expected;
    
    /**
     * Constructor.
     * @param a1 First component of the first triple.
     * @param a2 Second component of the first triple.
     * @param a3 Third component of the first triple.
     * @param b1 First component of the second triple.
     * @param b2 Second component of the second triple.
     * @param b3 Third component of the second triple.
     * @param expected Expected comparison result.
     */
    public CompareTriplesTest(final Integer a1,
                              final Integer a2,
                              final Integer a3,
                              final Integer b1,
                              final Integer b2,
                              final Integer b3,
                              final int expected) {
      this.a1 = a1;
      this.a2 = a2;
      this.a3 = a3;
      this.b1 = b1;
      this.b2 = b2;
      this.b3 = b3;
      this.expected = expected;
    }
    
    // Construct all possible pairs of triples with elements (null, 1, 2)
    // Use string comparison to check the result (see tupleToString()).
    @Parameters
    public static Collection<Object[]> data() {
      final Integer[] values= { null, 1, 2 };
      
      final ArrayList<Object[]> triples = new ArrayList<Object[]>();
      for (final Integer x1 : values) {
        for (final Integer x2 : values) {
          for (final Integer x3 : values) {
            triples.add(new Object[]{ x1, x2, x3});
          }
        }
      }
      
      final ArrayList<Object[]> data = new ArrayList<Object[]>();
      for (final Object[] first : triples) {
        for (final Object[] second : triples) {
          final int expected = 
              tupleToString(first).compareTo(tupleToString(second));
          data.add(new Object[]{ first[0], first[1], first[2], 
                                 second[0], second[1], second[2],
                                 expected});
        }
      }
      return data;
    }
    
    /** Comparison result is correct. */
    @Test
    public void testCompare() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertEquals(
         String.format(
            "Wrong comparison result for data: (%s, %s, %s), (%s, %s, %s)", 
            a1, a2, a3, b1, b2, b3),
         expected,
         cmpUtil.compare(a1, a2, a3, b1, b2, b3));
    }
  }  


  /** Test compare() on tuples of four (possibly null) elements. */
  @RunWith(Parameterized.class)
  public static class Compare4TuplesTest {
    /** First component of the first tuple. */
    private final Integer a1;
    
    /** Second component of the first tuple. */
    private final Integer a2;
    
    /** Third component of the first tuple. */
    private final Integer a3;

    /** Fourth component of the first tuple. */
    private final Integer a4;
    
    /** First component of the second tuple. */
    private final Integer b1;
    
    /** Second component of the second tuple. */
    private final Integer b2;

    /** Third component of the second tuple. */
    private final Integer b3;

    /** Fourth component of the second tuple. */
    private final Integer b4;
    
    /** Expected comparison result. */
    private final int expected;
    
    /**
     * Constructor.
     * @param a1 First component of the first tuple.
     * @param a2 Second component of the first tuple.
     * @param a3 Third component of the first tuple.
     * @param a4 Fourth component of the first tuple.
     * @param b1 First component of the second tuple.
     * @param b2 Second component of the second tuple.
     * @param b3 Third component of the second tuple.
     * @param b4 Fourth component of the second tuple.
     * @param expected Expected comparison result.
     */
    public Compare4TuplesTest(final Integer a1,
                              final Integer a2,
                              final Integer a3,
                              final Integer a4,
                              final Integer b1,
                              final Integer b2,
                              final Integer b3,
                              final Integer b4,
                              final int expected) {
      this.a1 = a1;
      this.a2 = a2;
      this.a3 = a3;
      this.a4 = a4;
      this.b1 = b1;
      this.b2 = b2;
      this.b3 = b3;
      this.b4 = b4;
      this.expected = expected;
    }
    
    // Construct all possible tuples with elements (1, null)
    // and all possible tuples with elements (1, 2).
    @Parameters
    public static Collection<Object[]> data() {
      final ArrayList<Object[]> data = new ArrayList<Object[]>();
      Integer[] values = { null, 1 };
      
      ArrayList<Object[]> tuples = new ArrayList<Object[]>();
      for (final Integer x1 : values) {
        for (final Integer x2 : values) {
          for (final Integer x3 : values) {
            for (final Integer x4 : values) {
              tuples.add(new Object[]{ x1, x2, x3, x4});
            }
          }
        }
      }
      
      for (final Object[] first : tuples) {
        for (final Object[] second : tuples) {
          final int expected = 
              tupleToString(first).compareTo(tupleToString(second));
          data.add(new Object[]{ first[0], first[1], first[2], first[3],
                                 second[0], second[1], second[2], second[3],
                                 expected});
        }
      }
      
      values = new Integer[]{ 1, 2 };
      tuples = new ArrayList<Object[]>();
      for (final Integer x1 : values) {
        for (final Integer x2 : values) {
          for (final Integer x3 : values) {
            for (final Integer x4 : values) {
              tuples.add(new Object[]{ x1, x2, x3, x4 });
            }
          }
        }
      }
      
      for (final Object[] first : tuples) {
        for (final Object[] second : tuples) {
          final int expected = 
              tupleToString(first).compareTo(tupleToString(second));
          data.add(new Object[]{ first[0], first[1], first[2], first[3],
                                 second[0], second[1], second[2], second[3],
                                 expected});
        }
      }
      
      return data;
    }
    
    /** Comparison result is correct. */
    @Test
    public void testCompare() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertEquals(
         String.format(
            "Wrong comparison result for data: (%s, %s, %s, %s), (%s, %s, %s, %s)", 
            a1, a2, a3, a4, b1, b2, b3, b4),
         expected,
         cmpUtil.compare(a1, a2, a3, a4, b1, b2, b3, b4));
    }
  }

  /** Test compare() on arrays of (possibly null) elements. */
  @RunWith(Parameterized.class)
  public static class CompareArraysTest {
    /** First array. */
    private final Integer[] first;
    
    /** Second array. */
    private final Integer[] second;
    
    /** Expected comparison result. */
    private final int expected;
    
    /**
     * Constructor.
     * @param first First array.
     * @param second Second array.
     * @param expected Expected comparison result.
     */
    public CompareArraysTest(final Integer[] first,
                             final Integer[] second,
                             final int expected) {
      this.first = first;
      this.second = second;
      this.expected = expected;
    }
    
    @Parameters
    public static Collection<Object[]> data() {
      final Object[][] data = new Object[][] {
        { null, null, 0 },
        { null, new Integer[]{}, -1 },
        { new Integer[]{}, null, 1 },
        { new Integer[]{}, new Integer[]{}, 0 },
        { new Integer[]{ 1, 2, 3 }, new Integer[]{ 1, 2 }, 1 },
        { new Integer[]{ 1, 2 }, new Integer[]{ 1, 2, 3 }, -1 },
        { new Integer[]{ null, null }, new Integer[]{ null, null }, 0 }, 
        { new Integer[]{ null, null }, new Integer[]{ null, 1 }, -1 }, 
        { new Integer[]{ null, null }, new Integer[]{ null, 2 }, -1 }, 
        { new Integer[]{ null, null }, new Integer[]{ 1, null }, -1 }, 
        { new Integer[]{ null, null }, new Integer[]{ 1, 1 }, -1 }, 
        { new Integer[]{ null, null }, new Integer[]{ 1, 2 }, -1 }, 
        { new Integer[]{ null, null }, new Integer[]{ 2, null }, -1 }, 
        { new Integer[]{ null, null }, new Integer[]{ 2, 1 }, -1 }, 
        { new Integer[]{ null, null }, new Integer[]{ 2, 2 }, -1 }, 
        { new Integer[]{ null, 1 }, new Integer[]{ null, null }, 1 }, 
        { new Integer[]{ null, 1 }, new Integer[]{ null, 1 }, 0 }, 
        { new Integer[]{ null, 1 }, new Integer[]{ null, 2 }, -1 }, 
        { new Integer[]{ null, 1 }, new Integer[]{ 1, null }, -1 }, 
        { new Integer[]{ null, 1 }, new Integer[]{ 1, 1 }, -1 }, 
        { new Integer[]{ null, 1 }, new Integer[]{ 1, 2 }, -1 }, 
        { new Integer[]{ null, 1 }, new Integer[]{ 2, null }, -1 }, 
        { new Integer[]{ null, 1 }, new Integer[]{ 2, 1 }, -1 }, 
        { new Integer[]{ null, 1 }, new Integer[]{ 2, 2 }, -1 }, 
        { new Integer[]{ null, 2 }, new Integer[]{ null, null }, 1 }, 
        { new Integer[]{ null, 2 }, new Integer[]{ null, 1 }, 1 }, 
        { new Integer[]{ null, 2 }, new Integer[]{ null, 2 }, 0 }, 
        { new Integer[]{ null, 2 }, new Integer[]{ 1, null }, -1 }, 
        { new Integer[]{ null, 2 }, new Integer[]{ 1, 1 }, -1 }, 
        { new Integer[]{ null, 2 }, new Integer[]{ 1, 2 }, -1 }, 
        { new Integer[]{ null, 2 }, new Integer[]{ 2, null }, -1 }, 
        { new Integer[]{ null, 2 }, new Integer[]{ 2, 1 }, -1 }, 
        { new Integer[]{ null, 2 }, new Integer[]{ 2, 2 }, -1 }, 
        { new Integer[]{ 1, null }, new Integer[]{ null, null }, 1 }, 
        { new Integer[]{ 1, null }, new Integer[]{ null, 1 }, 1 }, 
        { new Integer[]{ 1, null }, new Integer[]{ null, 2 }, 1 }, 
        { new Integer[]{ 1, null }, new Integer[]{ 1, null }, 0 }, 
        { new Integer[]{ 1, null }, new Integer[]{ 1, 1 }, -1 }, 
        { new Integer[]{ 1, null }, new Integer[]{ 1, 2 }, -1 }, 
        { new Integer[]{ 1, null }, new Integer[]{ 2, null }, -1 }, 
        { new Integer[]{ 1, null }, new Integer[]{ 2, 1 }, -1 }, 
        { new Integer[]{ 1, null }, new Integer[]{ 2, 2 }, -1 }, 
        { new Integer[]{ 1, 1 }, new Integer[]{ null, null }, 1 }, 
        { new Integer[]{ 1, 1 }, new Integer[]{ null, 1 }, 1 }, 
        { new Integer[]{ 1, 1 }, new Integer[]{ null, 2 }, 1 }, 
        { new Integer[]{ 1, 1 }, new Integer[]{ 1, null }, 1 }, 
        { new Integer[]{ 1, 1 }, new Integer[]{ 1, 1 }, 0 }, 
        { new Integer[]{ 1, 1 }, new Integer[]{ 1, 2 }, -1 }, 
        { new Integer[]{ 1, 1 }, new Integer[]{ 2, null }, -1 }, 
        { new Integer[]{ 1, 1 }, new Integer[]{ 2, 1 }, -1 }, 
        { new Integer[]{ 1, 1 }, new Integer[]{ 2, 2 }, -1 }, 
        { new Integer[]{ 1, 2 }, new Integer[]{ null, null }, 1 }, 
        { new Integer[]{ 1, 2 }, new Integer[]{ null, 1 }, 1 }, 
        { new Integer[]{ 1, 2 }, new Integer[]{ null, 2 }, 1 }, 
        { new Integer[]{ 1, 2 }, new Integer[]{ 1, null }, 1 }, 
        { new Integer[]{ 1, 2 }, new Integer[]{ 1, 1 }, 1 }, 
        { new Integer[]{ 1, 2 }, new Integer[]{ 1, 2 }, 0 }, 
        { new Integer[]{ 1, 2 }, new Integer[]{ 2, null }, -1 }, 
        { new Integer[]{ 1, 2 }, new Integer[]{ 2, 1 }, -1 }, 
        { new Integer[]{ 1, 2 }, new Integer[]{ 2, 2 }, -1 }, 
        { new Integer[]{ 2, null }, new Integer[]{ null, null }, 1 }, 
        { new Integer[]{ 2, null }, new Integer[]{ null, 1 }, 1 }, 
        { new Integer[]{ 2, null }, new Integer[]{ null, 2 }, 1 }, 
        { new Integer[]{ 2, null }, new Integer[]{ 1, null }, 1 }, 
        { new Integer[]{ 2, null }, new Integer[]{ 1, 1 }, 1 }, 
        { new Integer[]{ 2, null }, new Integer[]{ 1, 2 }, 1 }, 
        { new Integer[]{ 2, null }, new Integer[]{ 2, null }, 0 }, 
        { new Integer[]{ 2, null }, new Integer[]{ 2, 1 }, -1 }, 
        { new Integer[]{ 2, null }, new Integer[]{ 2, 2 }, -1 }, 
        { new Integer[]{ 2, 1 }, new Integer[]{ null, null }, 1 }, 
        { new Integer[]{ 2, 1 }, new Integer[]{ null, 1 }, 1 }, 
        { new Integer[]{ 2, 1 }, new Integer[]{ null, 2 }, 1 }, 
        { new Integer[]{ 2, 1 }, new Integer[]{ 1, null }, 1 }, 
        { new Integer[]{ 2, 1 }, new Integer[]{ 1, 1 }, 1 }, 
        { new Integer[]{ 2, 1 }, new Integer[]{ 1, 2 }, 1 }, 
        { new Integer[]{ 2, 1 }, new Integer[]{ 2, null }, 1 }, 
        { new Integer[]{ 2, 1 }, new Integer[]{ 2, 1 }, 0 }, 
        { new Integer[]{ 2, 1 }, new Integer[]{ 2, 2 }, -1 }, 
        { new Integer[]{ 2, 2 }, new Integer[]{ null, null }, 1 }, 
        { new Integer[]{ 2, 2 }, new Integer[]{ null, 1 }, 1 }, 
        { new Integer[]{ 2, 2 }, new Integer[]{ null, 2 }, 1 }, 
        { new Integer[]{ 2, 2 }, new Integer[]{ 1, null }, 1 }, 
        { new Integer[]{ 2, 2 }, new Integer[]{ 1, 1 }, 1 }, 
        { new Integer[]{ 2, 2 }, new Integer[]{ 1, 2 }, 1 }, 
        { new Integer[]{ 2, 2 }, new Integer[]{ 2, null }, 1 }, 
        { new Integer[]{ 2, 2 }, new Integer[]{ 2, 1 }, 1 }, 
        { new Integer[]{ 2, 2 }, new Integer[]{ 2, 2 }, 0 }                                
      };
      return Arrays.asList(data);
    }
    
    /** Comparison result is correct for non-null references. */
    @Test
    public void testCompare() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertEquals(
         String.format(
            "Wrong comparison result for data: %s, %s", 
            first, second),
         expected,
         cmpUtil.compare(first, second));
    }
  }


  /** Test compare() on iterables of (possibly null) elements. */
  @RunWith(Parameterized.class)
  public static class CompareIterablesTest {
    /** First array. */
    private final Iterable<Integer> first;
    
    /** Second array. */
    private final Iterable<Integer> second;
    
    /** Expected comparison result. */
    private final int expected;
    
    /**
     * Constructor.
     * @param first First array.
     * @param second Second array.
     * @param expected Expected comparison result.
     */
    public CompareIterablesTest(final Iterable<Integer> first,
                                final Iterable<Integer> second,
                                final int expected) {
      this.first = first;
      this.second = second;
      this.expected = expected;
    }
    
    /* Steal tests from array comparison test class. */
    @Parameters
    public static Collection<Object[]> data() {
      final Collection<Object[]> arrayData = CompareArraysTest.data();
      final List<Object[]> data = new ArrayList<Object[]>(arrayData.size());
      for (final Object[] row : arrayData) {
        final Iterable<Integer> it1;
        final Iterable<Integer> it2;
        
        if (row[0] == null) {
          it1 = null;
        } else {
          it1 = Arrays.asList((Integer[])row[0]);
        }
        
        if (row[1] == null) {
          it2 = null;
        } else {
          it2 = Arrays.asList((Integer[])row[1]);
        }
        
        data.add(new Object[] { it1, it2, row[2] });
      }
      return data;
    }
    
    /** Comparison result is correct for non-null references. */
    @Test
    public void testCompare() {
      final CmpUtil cmpUtil = CmpUtil.getInstance();
      assumeNotNull("Null returned from getInstance()", cmpUtil);
      assertEquals(
         String.format(
            "Wrong comparison result for data: %s, %s", 
            first, second),
         expected,
         cmpUtil.compare(first, second));
    }
  }
}
