package net.afyre.haha.demo.util.tests;

import static org.junit.Assert.*;
import static net.afyre.haha.demo.util.tests.Util.*;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import net.afyre.haha.demo.util.Function;
import net.afyre.haha.demo.util.MappedIterator;

import org.junit.Test;


/**
 * Tests for the {@link MappedIterator} class.
 * Test cases defined here are usually duplicated
 * in MappedIterableTest.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 * @see MappedIterableTest
 */
public class MappedIteratorTest {
  /** GetInstance does not return null. */
  @Test
  public void testInstanceNotNull() {
    final Integer[] array = { 1, 2, 3, 4 };
    final Iterator<Integer> source = Arrays.asList(array).iterator();
    final Function<Integer, Integer> function = new Inc();
    final Iterator<Integer> it = 
        MappedIterator.getInstance(source, function);
    assertNotNull("Null returned from getInstance", it);
  }

  /** NPE is thrown for null source. */
  @Test(expected=NullPointerException.class)
  public void testNullSource() {
    MappedIterator.getInstance(null, new Inc());
  }
  
  /** NPE is thrown for null function. */
  @Test(expected=NullPointerException.class)
  public void testNullFunction() {    
    final Integer[] array = { 1, 2, 3, 4 };
    final Iterator<Integer> source = Arrays.asList(array).iterator();
    MappedIterator.getInstance(source, null);
  }
  
  /** Mapping is applied correctly. */
  @Test
  public void testSimpleMap() {
    final Integer[] array = { 2, 1, 4, 3 };
    final Iterator<Integer> source = Arrays.asList(array).iterator();
    final Function<Integer, Integer> function = new Inc();
    final Iterator<Integer> it = 
        MappedIterator.getInstance(source, function);
    assertIteratorReturns(it, 3, 2, 5, 4);
  }
  
  /** Iterator works correctly for empty sequences. */
  @Test
  public void testEmpty() {
    final Integer[] array = { };
    final Iterator<Integer> source = Arrays.asList(array).iterator();
    final Function<Integer, Integer> function = new Inc();
    final Iterator<Integer> it = 
        MappedIterator.getInstance(source, function);
    assertIteratorReturns(it);    
  }
  
  /** Remove() requests are forwarded. */
  @Test
  public void testRemove() {
    final List<Integer> sourceList = new ArrayList<Integer>();
    sourceList.addAll(Arrays.asList(1, 2, 3, 4, 5, 6));
    final Iterator<Integer> source = sourceList.iterator();
    final Function<Integer, Integer> function = new Inc();
    final Iterator<Integer> it = 
        MappedIterator.getInstance(source, function);
    it.next();
    it.next();
    it.next();
    it.remove();
    assertArrayEquals("Wrong values after remove()",
                      sourceList.toArray(), 
                      new Object[]{ 1, 2, /*3,*/ 4, 5, 6 });
  }
  
  private static class Inc implements Function<Integer, Integer> {
    /** {@inheritDoc} */
    @Override
    public Integer apply(final Integer arg) {
      return arg + 1;
    }
  }
  
}
