package net.afyre.haha.demo.util.tests;

import static org.junit.Assert.*;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.jmock.Expectations;
import org.jmock.Mockery;
import org.jmock.integration.junit4.JUnit4Mockery;
import org.jmock.lib.legacy.ClassImposteriser;
import org.junit.Before;
import org.junit.Test;

import net.afyre.haha.demo.util.TeeInputStream;

/**
 * Test for the {@link TeeInputStream} class.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public class TeeInputStreamTest {
  /** JMock context. */
  Mockery mockery;
  
  /** Create JMock context. */
  @Before
  public void setUp() {
    mockery = new JUnit4Mockery() {{
      setImposteriser(ClassImposteriser.INSTANCE);
    }};
  }
  
  /** Exception is thrown when a null output is passed to getInstance. */
  @Test(expected=NullPointerException.class)
  public void testGetInstanceNullOutput() {
    final InputStream mock = mockery.mock(InputStream.class);
    TeeInputStream.getInstance(mock, null, true);
  }

  /** Exception is thrown when a null input is passed to getInstance. */
  @Test(expected=NullPointerException.class)
  public void testGetInstanceNullInput() {
    final OutputStream mock = mockery.mock(OutputStream.class);
    TeeInputStream.getInstance(null, mock, true);
  }

  /** Output stream is closed on close() call if requested. */
  @Test
  public void testCloseTrue() throws IOException {
    final OutputStream outputMock = mockery.mock(OutputStream.class);
    final InputStream inputMock = mockery.mock(InputStream.class);
    
    // Expectations
    mockery.checking(new Expectations() {{
      oneOf(inputMock).close();
      oneOf(outputMock).close();
    }});
    
    // Execute test
    final TeeInputStream stream = 
        TeeInputStream.getInstance(inputMock, outputMock, true);
    
    stream.close();     
  }  

  /** Output stream is not closed on close() call if requested. */
  @Test
  public void testCloseFalse() throws IOException {
    final OutputStream outputMock = mockery.mock(OutputStream.class);
    final InputStream inputMock = mockery.mock(InputStream.class);
    
    // Expectations
    mockery.checking(new Expectations() {{
      oneOf(inputMock).close();
    }});
    
    // Execute test
    final TeeInputStream stream = 
        TeeInputStream.getInstance(inputMock, outputMock, false);
    
    stream.close();     
  }  

  /** Read() causes a write to the output stream. */
  @Test
  public void testReadInt() throws IOException {
    final OutputStream outputMock = mockery.mock(OutputStream.class);
    final InputStream inputMock = mockery.mock(InputStream.class);
    
    // Expectations
    mockery.checking(new Expectations() {{
      oneOf(inputMock).read();
      will(returnValue(42));
      
      oneOf(outputMock).write(42);
    }});
    
    // Execute test
    final TeeInputStream stream = 
        TeeInputStream.getInstance(inputMock, outputMock, false);
    
    final int result = stream.read();
    assertEquals("Wrong value returned from read", 42, result);
  }
  
  /** Read(byte[], int, int) causes a write to the output stream. */
  @Test
  public void testReadByteWithOffset() throws IOException {
    final OutputStream outputMock = mockery.mock(OutputStream.class);
    final InputStream inputMock = mockery.mock(InputStream.class);
    final byte[] data = { 100, 117, 112, 97 };
    
    // Expectations
    mockery.checking(new Expectations() {{
      oneOf(inputMock).read(data, 2, 1);
      will(returnValue(1));
      
      oneOf(outputMock).write(data, 2, 1);
    }});
    
    // Execute test
    final TeeInputStream stream = 
        TeeInputStream.getInstance(inputMock, outputMock, false);
    

    final int result = stream.read(data, 2, 1);
    assertEquals("Wrong value returned from read", 1, result);  
  }

  /** Read(byte[]) causes a write to the output stream. */
  @Test
  public void testReadByte() throws IOException {
    final OutputStream outputMock = mockery.mock(OutputStream.class);
    final InputStream inputMock = mockery.mock(InputStream.class);
    final byte[] data = { 100, 117, 112, 97 };
    
    // Expectations
    mockery.checking(new Expectations() {{
      oneOf(inputMock).read(data);
      will(returnValue(4));
      
      oneOf(outputMock).write(data);
    }});
    
    // Execute test
    final TeeInputStream stream = 
        TeeInputStream.getInstance(inputMock, outputMock, false);
    
    final int result = stream.read(data);
    assertEquals("Wrong value returned from read", 4, result);    
  }
  
  /** Read() causes no write on error. */
  @Test
  public void testReadIntError() throws IOException {
    final OutputStream outputMock = mockery.mock(OutputStream.class);
    final InputStream inputMock = mockery.mock(InputStream.class);
    
    // Expectations
    mockery.checking(new Expectations() {{
      oneOf(inputMock).read();
      will(returnValue(-1));
    }});
    
    // Execute test
    final TeeInputStream stream = 
        TeeInputStream.getInstance(inputMock, outputMock, false);
    
    final int result = stream.read();
    assertEquals("Wrong value returned from read", -1, result);    
  }
  
  /** Read(byte[], int, int) causes no write on error. */
  @Test
  public void testReadByteWithOffsetError() throws IOException {
    final OutputStream outputMock = mockery.mock(OutputStream.class);
    final InputStream inputMock = mockery.mock(InputStream.class);
    final byte[] data = { 100, 117, 112, 97 };
    
    // Expectations
    mockery.checking(new Expectations() {{
      oneOf(inputMock).read(data, 2, 1);
      will(returnValue(-1));
    }});
    
    // Execute test
    final TeeInputStream stream = 
        TeeInputStream.getInstance(inputMock, outputMock, false);
    
    final int result = stream.read(data, 2, 1);
    assertEquals("Wrong value returned from read", -1, result);  
  }

  /** Read(byte[]) causes no write on error. */
  @Test
  public void testReadByteError() throws IOException {
    final OutputStream outputMock = mockery.mock(OutputStream.class);
    final InputStream inputMock = mockery.mock(InputStream.class);
    final byte[] data = { 100, 117, 112, 97 };
    
    // Expectations
    mockery.checking(new Expectations() {{
      oneOf(inputMock).read(data);
      will(returnValue(-1));
    }});
    
    // Execute test
    final TeeInputStream stream = 
        TeeInputStream.getInstance(inputMock, outputMock, false);
    
    final int result = stream.read(data);
    assertEquals("Wrong value returned from read", -1, result);       
  }
}
