/**
 * 
 */
package net.afyre.haha.demo.util;

import java.util.Iterator;

/**
 * Iterable wrapper for arrays.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 * @type <T> Type of array elements.
 */
public final class ArrayIterable<T> implements Iterable<T> {
  /** Source array. */
  private final T[] array;
  
  /** Index of the first element that will be returned. */
  private int firstIndex;
  
  /** Index of the last element that will be returned. */
  private final int lastIndex;
  
  /**
   * Constructor.
   * @param array Source array.
   * @throws NullPointerException If the array reference is null.
   */
  private ArrayIterable(final T[] array) {
    if (array == null) {
      throw new NullPointerException("array");
    }
    this.array = array;
    this.firstIndex = 0;
    this.lastIndex = array.length - 1;
  }
  
  /**
   * Constructor.
   * @param array Source array.
   * @param startIndex Index of the first element to be returned.
   * @throws NullPointerException If the array reference is null.
   */
  private ArrayIterable(final T[] array, 
                        final int firstIndex) {
    if (array == null) {
      throw new NullPointerException("array");
    }
    this.array = array;
    this.firstIndex = firstIndex;
    this.lastIndex = array.length - 1;
  }
    
  /**
   * Constructor.
   * @param array Source array.
   * @param startIndex Index of the first element to be returned.
   * @param lastIndex Index of the last element to be returned.
   * @throws NullPointerException If the array reference is null.
   */
  private ArrayIterable(final T[] array, 
                        final int firstIndex, 
                        final int lastIndex) {
    if (array == null) {
      throw new NullPointerException("array");
    }
    this.array = array;
    this.firstIndex = firstIndex;
    this.lastIndex = lastIndex;
  }
  
  
  /**
   * Wrap an array.
   * @param <T> Type of elements.
   * @param array Source array.
   * @param startIndex Index of the first element to be returned.
   * @param lastIndex Index of the last element to be returned.
   * @return New iterator.
   * @throws NullPointerException If the array reference is null.
   */
  public static<T> Iterable<T> getInstance(final T[] array, 
                                           final int firstIndex,
                                           final int lastIndex) {
    return new ArrayIterable<T>(array, firstIndex, lastIndex);
  }
  
  /**
   * Wrap a fragment of an array.
   * @param <T> Type of elements.
   * @param array Source array.
   * @param startIndex Index of the first element to be returned.
   * @return New iterator.
   * @throws NullPointerException If the array reference is null.
   */
  public static<T> Iterable<T> getInstance(final T[] array, 
                                           final int firstIndex) {
    return new ArrayIterable<T>(array, firstIndex);
  }
  
  /**
   * Wrap a fragment of an array.
   * @param <T> Type of elements.
   * @param array Source array.
   * @return New iterator.
   * @throws NullPointerException If the array reference is null.
   */
  public static<T> Iterable<T> getInstance(final T[] array) {
    return new ArrayIterable<T>(array);
  }
  
  /** {@inheritDoc} */
  @Override
  public Iterator<T> iterator() {
    return ArrayIterator.getInstance(array, firstIndex, lastIndex);
  }

}
