/**
 * 
 */
package net.afyre.haha.demo.util;

import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * An iterator returning elements of an array.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 * @param <T> Type of array elements.
 */
public final  class ArrayIterator<T> implements Iterator<T> {
  /** Source array. */
  private final T[] array;
  
  /** Index of the next element that will be returned. */
  private int currentIndex;
  
  /** Index of the last element that will be returned. */
  private final int lastIndex;
  
  /**
   * Constructor.
   * @param array Source array.
   * @throws NullPointerException If the array reference is null.
   */
  private ArrayIterator(final T[] array) {
    if (array == null) {
      throw new NullPointerException("array");
    }
    this.array = array;
    this.currentIndex = 0;
    this.lastIndex = array.length - 1;
  }
  
  /**
   * Constructor.
   * @param array Source array.
   * @param startIndex Index of the first element to be returned.
   *                   If it is greater than array.length, 
   *                   the resulting iterator is empty (but valid).
   * @throws NullPointerException If the array reference is null.
   */
  private ArrayIterator(final T[] array, 
                        final int firstIndex) {
    if (array == null) {
      throw new NullPointerException("array");
    }
    this.array = array;
    this.currentIndex = firstIndex;
    this.lastIndex = array.length - 1;
  }
    
  /**
   * Constructor.
   * @param array Source array.
   * @param startIndex Index of the first element to be returned.
   *                   If it is greater than lastIndex, 
   *                   the resulting iterator is empty (but valid).
   * @param lastIndex Index of the last element to be returned.
   *                  Can be greater than array.length.
   * @throws NullPointerException If the array reference is null.
   */
  private ArrayIterator(final T[] array, 
                        final int firstIndex, 
                        final int lastIndex) {
    if (array == null) {
      throw new NullPointerException("array");
    }
    this.array = array;
    this.currentIndex = firstIndex;
    this.lastIndex = 
        lastIndex < array.length ? lastIndex : array.length - 1;
  }
  
  /**
   * Create an iterator over elements of an array.
   * @param <T> Type of elements.
   * @param array Source array.
   * @param startIndex Index of the first element to be returned.
   *                   If it is greater than lastIndex, 
   *                   the resulting iterator is empty (but valid).
   * @param lastIndex Index of the last element to be returned.
   *                  Can be greater than array.length.
   * @return New iterator.
   * @throws NullPointerException If the array reference is null.
   */
  public static<T> Iterator<T> getInstance(final T[] array, 
                                           final int firstIndex,
                                           final int lastIndex) {
    return new ArrayIterator<T>(array, firstIndex, lastIndex);
  }
  
  /**
   * Create an iterator over elements of an array.
   * @param <T> Type of elements.
   * @param array Source array.
   * @param startIndex Index of the first element to be returned.
   *                   If it is greater than array.length, 
   *                   the resulting iterator is empty (but valid).
   * @return New iterator.
   * @throws NullPointerException If the array reference is null.
   */
  public static<T> Iterator<T> getInstance(final T[] array, 
                                           final int firstIndex) {
    return new ArrayIterator<T>(array, firstIndex);
  }
  
  /**
   * Create an iterator over elements of an array.
   * @param <T> Type of elements.
   * @param array Source array.
   * @return New iterator.
   * @throws NullPointerException If the array reference is null.
   */
  public static<T> Iterator<T> getInstance(final T[] array) {
    return new ArrayIterator<T>(array);
  }
  
  /** {@inheritDoc} */
  @Override
  public boolean hasNext() {
    return currentIndex <= lastIndex;
  }

  /** {@inheritDoc} */
  @Override
  public T next() {
    if (currentIndex > lastIndex) {
      throw new NoSuchElementException();
    }
    return array[currentIndex++];
  }

  /** {@inheritDoc} */
  @Override
  public void remove() {
    throw new UnsupportedOperationException();
  }

}
