/**
 * 
 */
package net.afyre.haha.demo.util;

import java.util.Iterator;

/**
 * An iterator wrapper which applies a transformation
 * to each element returned.
 * 
 * Element removal is supported if the source iterator supports it.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 * @param <Src> Type of elements returned by the original iterator.
 * @param <T> Type of elements returned by this iterator.
 */
public final class MappedIterator<Src, T> implements Iterator<T> {
  /** Source iterator. */
  private final Iterator<? extends Src> source;
  
  /** Mapping function. */
  private final Function<? super Src, ? extends T> function;
  
  /**
   * Constructor.
   * @param source Source iterator.
   * @param function Mapping function.
   * @throws NullPointerException If either of the arguments is null.
   */
  private MappedIterator(
      final Iterator<? extends Src> source,
      final Function<? super Src, ? extends T> function) {
    if (source == null) {
      throw new NullPointerException("source");
    }
    if (function == null) {
      throw new NullPointerException("function");
    }
    this.source = source;
    this.function = function;
  }
  
  /**
   * Create a mapped iterator.
   * @param <Src> Type of elements returned by the source iterator.
   * @param <T> Type of elements returned by this iterator.
   * @param source Source iterator.
   * @param function Mapping function.
   * @throws NullPointerException If either of the arguments is null.
   */
  public static<Src, T> Iterator<T> getInstance(
      final Iterator<? extends Src> source,
      final Function<? super Src, ? extends T> function) {
    return new MappedIterator<Src, T>(source, function);
  }

  /** {@inheritDoc} */
  @Override
  public boolean hasNext() {
    return source.hasNext();
  }

  /** {@inheritDoc} */
  @Override
  public T next() {
    return function.apply(source.next());
  }

  /** {@inheritDoc} */
  @Override
  public void remove() {
    source.remove();
  }
}
