package net.afyre.haha.demo.vcgen;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.ecore.EObject;

import com.google.common.collect.ImmutableList;

import net.afyre.haha.demo.haha.Argument;
import net.afyre.haha.demo.haha.Axiom;
import net.afyre.haha.demo.haha.Function;
import net.afyre.haha.demo.haha.Local;
import net.afyre.haha.demo.haha.Locals;
import net.afyre.haha.demo.haha.Postcondition;
import net.afyre.haha.demo.haha.Precondition;
import net.afyre.haha.demo.haha.Predicate;
import net.afyre.haha.demo.haha.Program;
import net.afyre.haha.demo.solver.context.AxiomDeclaration;
import net.afyre.haha.demo.solver.context.Declaration;
import net.afyre.haha.demo.solver.context.FunctionDeclaration;
import net.afyre.haha.demo.solver.context.PredicateDeclaration;
import net.afyre.haha.demo.solver.context.TermDeclaration;
import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.types.ValueType;

/**
 * Verification condition generator for functions and programs.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 */
public final class HoareVCGen {
  /** VCGen for statements. */
  private final DefaultHoareTripleVCGen statementVCGen;

//  /** Utility methods. */
//  private final VCGenUtil util = VCGenUtil.getInstance();
  
  /** Object used to convert AST expressions to terms. */
  private final ExpressionUtil exprUtil = ExpressionUtil.getInstance();
  
//  /** Object used to create expressions. */
//  private final ExpressionFactory exprFactory = 
//      new SimpleExpressionFactory();
  
  /** Object used to convert expressions to formulae. */
  private final ExprToDocumentedFormula exprToFormula = 
      new ExprToDocumentedFormula();
  
  /** Constructor. */
  public HoareVCGen() {
    statementVCGen = new DefaultHoareTripleVCGen();
  }
  
  /**
   * Generate verification conditions for a given program.
   * @param context Parent context for generated conditions.
   * @param program Program to be processed.
   * @param consumer Object used to process generated conditions.
   */
  public void generateVerificationConditions(
     final VCContext context,
     final Program program,
     final VCConsumer consumer) {
   
    final VCContext rootContext = 
        new VCContext("main", 
                      ImmutableList.<EObject>of(program),
                      "Program correctness",
                      context, getDeclarations(program));
    
    for (final Function function : program.getFunctions()) {
      generateVerificationConditions(rootContext, function, consumer);
    }
  }
  
  /**
   * Generate verification conditions for a given function.
   * @param context Parent context for generated conditions.
   * @param function Function to be processed.
   * @param consumer Object used to process generated conditions.
   */
  public void generateVerificationConditions(
     final VCContext context,
     final Function function,
     final VCConsumer consumer) {
    
    // Allow abstract functions
    if (function.getBody() == null) {
      return;
    }
    
    final VCContext rootContext = 
        new VCContext("correctness_" + function.getName(), 
                      ImmutableList.<EObject>of(function),
                      "Correctness of " + function.getName(),
                      context, getDeclarations(function));
        
    List<DocumentedFormula> pre = new ArrayList<DocumentedFormula>();
    List<DocumentedFormula> post = new ArrayList<DocumentedFormula>();
    
    for (final Precondition precond : function.getPreconditions()) {
      pre.add(exprToFormula.apply(precond.getFormula()));
    }
    for (final Postcondition postcond : function.getPostconditions()) {
      post.add(exprToFormula.apply(postcond.getFormula()));
    }
    
    statementVCGen.generateVerificationConditions(
        rootContext, pre, function.getBody(), post, consumer);
  }
  
  
  /**
   * Get list of global declarations for a given program.
   * @param program Program
   * @return List of declarations.
   */
  private List<Declaration> getDeclarations(final Program program) {
    final List<Declaration> declarations = new ArrayList<Declaration>();

    declarations.addAll(getPredicateDeclarations(program));
    declarations.addAll(getFunctionDeclarations(program));
    //declarations.addAll(getFunctionSpecDeclarations(program));
    declarations.addAll(getAxiomDeclarations(program));
    
    return declarations;
  }
    
  /**
   * Get list of declarations arising from axioms in a program.
   * @param program Program.
   * @return List of declarations.
   */
  private List<Declaration> getAxiomDeclarations(final Program program) {
    final List<Declaration> declarations = new ArrayList<Declaration>();
    
    int counter = 0;
    for (final Axiom axiom : program.getAxioms()) {
      final String name;
      if (axiom.getName() == null) {
        name = String.format("Axiom%d", counter++);
      } else {
        name = axiom.getName();
      }
      final Formula formula = exprUtil.translateFormula(axiom.getFormula());
      declarations.add(new AxiomDeclaration(name, formula));
    }
    
    return declarations;
  }
  
  /**
   * Get list of declarations arising from predicates in a program.
   * @param program Program.
   * @return List of declarations.
   */
  private List<Declaration> 
  getPredicateDeclarations(final Program program) {
    final List<Declaration> declarations = new ArrayList<Declaration>();
    
    for (final Predicate predicate : program.getPredicates()) {
      // Collect arguments
      final List<VariableBinding> args = 
          new ArrayList<VariableBinding>();
      for (final Argument arg : 
               predicate.getArguments().getArguments()) {
        final ValueType argType = exprUtil.translateType(arg.getType());
        args.add(new VariableBinding(arg.getName(), argType));
      }
    
      final Formula body = 
          exprUtil.translateFormula(predicate.getFormula());
      final Declaration predicateDecl = 
          new PredicateDeclaration(predicate.getName() +"?", args, body);
      declarations.add(predicateDecl);
    }  
    
    return declarations;
  }
  
  /**
   * Get list of declarations arising from functions in a program.
   * @param program Program.
   * @return List of declarations.
   */
  private List<Declaration> 
  getFunctionDeclarations(final Program program) {
    final List<Declaration> declarations = new ArrayList<Declaration>();
    
    for (final Function function : program.getFunctions()) {
      final List<VariableBinding> args = 
          new ArrayList<VariableBinding>();
      for (final Argument arg : 
               function.getArguments().getArguments()) {
        final ValueType argType = exprUtil.translateType(arg.getType());
        args.add(new VariableBinding(arg.getName(), argType));
      }      
      final ValueType resultType = 
          exprUtil.translateType(function.getResultType());
      final Declaration functionDecl = 
          new FunctionDeclaration("function@" + function.getName(), args, resultType);
      declarations.add(functionDecl);
    }
    
    return declarations;
  }

//  /**
//   * Get list of declarations arising from 
//   * function specifications in a program.
//   * 
//   * WARNING: UNSOUND - we need to check feasibility and guarantee termination.
//   * 
//   * @param program Program.
//   * @return List of declarations.
//   */
//  private List<Declaration> 
//  getFunctionSpecDeclarations(final Program program) {
//    final List<Declaration> declarations = new ArrayList<Declaration>();
//    
//    for (final Function function : program.getFunctions()) {
//      final Term call = util.getFunctionCall(function);
//      final Expression pattern = call;
//      // Used for unnamed postconditions
//      int counter = 0;
//      
//      for (final Postcondition post : function.getPostconditions()) {
//        Formula spec = exprUtil.translateFormula(post.getFormula());
//        final Substitution subst = new Substitution();
//        subst.setTermBinding(function.getName(), call);
//        spec = subst.perform(spec);
//        
//        for (final Precondition pre : function.getPreconditions()) {
//          final Formula preExpr = 
//              exprUtil.translateFormula(pre.getFormula());
//          spec = exprFactory.createImplication(preExpr, spec);
//        }
//        final List<VariableBinding> bindings = new ArrayList<VariableBinding>(
//            function.getArguments().getArguments().size());
//        for (final Argument arg : 
//                 function.getArguments().getArguments()) {
//          final ValueType argType = exprUtil.translateType(arg.getType());
//          bindings.add(new VariableBinding(arg.getName(), argType));
//        }
//        if (bindings.size() > 0) {
//          spec = exprFactory.createForall(bindings, spec, pattern);
//        }
//        final String name;
//        if (post.getName() != null) {
//          name = function.getName() + "--" + post.getName();
//        } else {
//          name = function.getName() + "-post-" + counter++;
//        }
//        final Declaration axiomDecl = new AxiomDeclaration(name, spec);
//        declarations.add(axiomDecl); 
//      }
//    }
//    
//    return declarations;
//  }
  
  /**
   * Get list of declarations corresponding to arguments
   * and locals of a given function.
   * @param function Function.
   * @return List of declarations.
   */
  private List<Declaration> getDeclarations(final Function function) {
    final List<Declaration> declarations = new ArrayList<Declaration>();
    
    // Arguments
    for (final Argument arg : function.getArguments().getArguments()) {
      final ValueType type = exprUtil.translateType(arg.getType());
      final Declaration argDecl = 
          new TermDeclaration(arg.getName(), type);
      declarations.add(argDecl);
    }
    
    // Result variable
    final ValueType resultType = 
        exprUtil.translateType(function.getResultType());
    // WARNING: This must be consistent with ExpressionSwitch#caseResultVar()
    final Declaration resultDecl = 
        new TermDeclaration("result-of-" + function.getName(), resultType);
    declarations.add(resultDecl);
    
    // Local variables
    for (final Locals locals : function.getLocals()) {
      for (final Local local : locals.getVariables()) {
        final ValueType type = exprUtil.translateType(local.getType());
        final Declaration localDecl =
            new TermDeclaration(local.getName(), type);
        declarations.add(localDecl);
      }
    }
    
    return declarations;
  }
  
}
