package net.afyre.haha.demo.vcgen.tree;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.afyre.haha.demo.vcgen.VCConsumer;
import net.afyre.haha.demo.vcgen.VCContext;
import net.afyre.haha.demo.vcgen.VerificationCondition;

/**
 * Builder used to create trees from verification conditions.
 * 
 * The builder acts as a {@link VCConsumer}. After processing
 * all conditions {@link #finish()} should be invoked to
 * recover the tree.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 */
public class TreeBuilder implements VCConsumer {
  private ContextNodeImpl root = new ContextNodeImpl(null, null);
  
  /** {@inheritDoc} */
  @Override
  public void processVerificationCondition(final VerificationCondition vc) {
    final List<VCContext> contexts = new ArrayList<VCContext>();
    
    for (VCContext context = vc.getContext(); 
         context != null; 
         context = context.getParent()) {
      contexts.add(context);
    }
    Collections.reverse(contexts);
    
    ContextNodeImpl node = root;
    for (final VCContext context : contexts) {
      node = node.getOrCreateChildContext(context);
    }
    node.addChildCondition(vc);
  }

  /**
   * Return current tree and reset the builder.
   * @return Root context node.
   */
  public ContextNode finish() {
    final ContextNode result = root;
    root = new ContextNodeImpl(null, null);
    return result;
  }
  
  private static final class ContextNodeImpl extends AbstractNode 
                                             implements ContextNode {
    private final VCContext context;
    
    private final Map<VCContext, ContextNodeImpl> contexts;
    private final List<Node> childNodes;
    
    public ContextNodeImpl(
        final Node parent,
        final VCContext context) {
      super(parent);
      this.context = context;
      childNodes = new ArrayList<Node>();
      contexts = new HashMap<VCContext, ContextNodeImpl>();
    }

    public ContextNodeImpl getOrCreateChildContext(final VCContext child) {
      final ContextNodeImpl result;
      
      if (!contexts.containsKey(child)) {
        result = new ContextNodeImpl(this, child);
        contexts.put(child, result);
        childNodes.add(result);
      } else {
        result = contexts.get(child);
      }
      return result;
    }
    
    public void addChildCondition(final VerificationCondition child) {
      childNodes.add(new ConditionNodeImpl(this, child));
    }
    
    /** {@inheritDoc} */
    @Override
    public <T> T doSwitch(NodeSwitch<? extends T> sw) {
      return sw.caseContext(this);
    }

    /** {@inheritDoc} */
    @Override
    public Collection<? extends Node> getChildren() {
      return Collections.unmodifiableList(childNodes);
    }

    /** {@inheritDoc} */
    @Override
    public VCContext getContext() {
      return context;
    }  
  }
  
  private static final class ConditionNodeImpl extends AbstractNode
                                               implements ConditionNode {
    private final VerificationCondition condition;
    
    public ConditionNodeImpl(final Node parent,
                             final VerificationCondition condition) {
      super(parent);
      this.condition = condition;
    }
    
    /** {@inheritDoc} */
    @Override
    public <T> T doSwitch(NodeSwitch<? extends T> sw) {
      return sw.caseCondition(this);
    }

    /** {@inheritDoc} */
    @Override
    public Collection<? extends Node> getChildren() {
      return Collections.emptyList();
    }

    /** {@inheritDoc} */
    @Override
    public VerificationCondition getCondition() {
      return condition;
    } 
  }
  
}
